# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
from io import StringIO

import pyomo.common.unittest as unittest

from pyomo.common.errors import IterationLimitError
from pyomo.common.log import LoggingIntercept
from pyomo.environ import (
    ConcreteModel,
    Var,
    Constraint,
    Param,
    ExternalFunction,
    value,
    exp,
    NonNegativeReals,
    Binary,
)
from pyomo.util.calc_var_value import calculate_variable_from_constraint
from pyomo.core.expr.calculus.diff_with_sympy import differentiate_available
from pyomo.core.expr.calculus.derivatives import differentiate
from pyomo.core.expr.sympy_tools import sympy_available

all_diff_modes = [
    differentiate.Modes.sympy,
    differentiate.Modes.reverse_symbolic,
    differentiate.Modes.reverse_numeric,
]


def sum_sq(args, fixed, fgh):
    f = sum(arg**2 for arg in args)
    g = [2 * arg for arg in args]
    h = None
    return f, g, h


class Test_calc_var(unittest.TestCase):
    def test_initialize_value(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var(initialize=0)
        m.c = Constraint(expr=m.x == 5)

        for mode in all_diff_modes:
            m.x.set_value(None)
            calculate_variable_from_constraint(m.x, m.c, diff_mode=mode)
            self.assertEqual(value(m.x), 5)

        m.x.setlb(3)
        for mode in all_diff_modes:
            m.x.set_value(None)
            calculate_variable_from_constraint(m.x, m.c, diff_mode=mode)
            self.assertEqual(value(m.x), 5)

        m.x.setlb(-10)
        for mode in all_diff_modes:
            m.x.set_value(None)
            calculate_variable_from_constraint(m.x, m.c, diff_mode=mode)
            self.assertEqual(value(m.x), 5)

        m.x.setub(10)
        for mode in all_diff_modes:
            m.x.set_value(None)
            calculate_variable_from_constraint(m.x, m.c, diff_mode=mode)
            self.assertEqual(value(m.x), 5)

        m.x.setlb(3)
        for mode in all_diff_modes:
            m.x.set_value(None)
            calculate_variable_from_constraint(m.x, m.c, diff_mode=mode)
            self.assertEqual(value(m.x), 5)

        m.x.setlb(None)
        for mode in all_diff_modes:
            m.x.set_value(None)
            calculate_variable_from_constraint(m.x, m.c, diff_mode=mode)
            self.assertEqual(value(m.x), 5)

        m.x.setub(-10)
        for mode in all_diff_modes:
            m.x.set_value(None)
            calculate_variable_from_constraint(m.x, m.c, diff_mode=mode)
            self.assertEqual(value(m.x), 5)

        m.lt = Constraint(expr=m.x <= m.y)
        with self.assertRaisesRegex(
            ValueError, "Constraint 'lt' must be an equality constraint"
        ):
            calculate_variable_from_constraint(m.x, m.lt)

        m.indexed = Constraint([1, 2], rule=lambda m, i: m.x <= i)
        with self.assertRaisesRegex(
            ValueError,
            r"calculate_variable_from_constraint\(\): constraint must be a scalar "
            r"constraint or a single ConstraintData.  Received IndexedConstraint "
            r'\("indexed"\)',
        ):
            calculate_variable_from_constraint(m.x, m.indexed)

    def test_linear(self):
        m = ConcreteModel()
        m.x = Var()
        m.c = Constraint(expr=5 * m.x == 10)

        for mode in all_diff_modes:
            m.x.set_value(None)
            calculate_variable_from_constraint(m.x, m.c, diff_mode=mode)
            self.assertEqual(value(m.x), 2)

    def test_constraint_as_tuple(self):
        m = ConcreteModel()
        m.x = Var()
        m.p = Param(initialize=15, mutable=True)

        for mode in all_diff_modes:
            m.x.set_value(None)
            calculate_variable_from_constraint(m.x, 5 * m.x == 5, diff_mode=mode)
            self.assertEqual(value(m.x), 1)
        for mode in all_diff_modes:
            m.x.set_value(None)
            calculate_variable_from_constraint(m.x, (5 * m.x, 10), diff_mode=mode)
            self.assertEqual(value(m.x), 2)
        for mode in all_diff_modes:
            m.x.set_value(None)
            calculate_variable_from_constraint(m.x, (15, 5 * m.x, m.p), diff_mode=mode)
            self.assertEqual(value(m.x), 3)
        with self.assertRaisesRegex(
            ValueError,
            "Constraint 'tuple' is a Ranged Inequality with a variable upper bound.",
        ):
            calculate_variable_from_constraint(m.x, (15, 5 * m.x, m.x))

    @unittest.skipIf(not differentiate_available, "this test requires sympy")
    def test_nonlinear(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var(initialize=0)

        m.c = Constraint(expr=m.x**2 == 16)
        for mode in all_diff_modes:
            m.x.set_value(1.0)  # set an initial value
            calculate_variable_from_constraint(
                m.x, m.c, linesearch=False, diff_mode=mode
            )
            self.assertEqual(value(m.x), 4)

        # test that infeasible constraint throws error
        m.d = Constraint(expr=m.x**2 == -1)
        for mode in all_diff_modes:
            m.x.set_value(1.25)  # set the initial value
            with self.assertRaisesRegex(
                IterationLimitError, r'Iteration limit \(10\) reached'
            ):
                calculate_variable_from_constraint(
                    m.x, m.d, iterlim=10, linesearch=False, diff_mode=mode
                )

        # same problem should throw a linesearch error if linesearch is on
        for mode in all_diff_modes:
            m.x.set_value(1.25)  # set the initial value
            with self.assertRaisesRegex(
                IterationLimitError, "Linesearch iteration limit reached"
            ):
                calculate_variable_from_constraint(
                    m.x, m.d, iterlim=10, linesearch=True, diff_mode=mode
                )

        # same problem should raise an error if initialized at 0
        for mode in all_diff_modes:
            m.x = 0
            with self.assertRaisesRegex(
                ValueError,
                "Initial value for variable 'x' results in a "
                "derivative value for constraint 'c' that is very close to zero.",
            ):
                calculate_variable_from_constraint(m.x, m.c, diff_mode=mode)

        # same problem should raise a value error if we are asked to
        # solve for a variable that is not present
        for mode in all_diff_modes:
            if mode == differentiate.Modes.reverse_numeric:
                # numeric differentiation should not be used to check if a
                # derivative is always zero
                with self.assertRaisesRegex(
                    ValueError,
                    "Initial value for variable 'y' results in a "
                    "derivative value for constraint 'c' that is very close to zero.",
                ):
                    calculate_variable_from_constraint(m.y, m.c, diff_mode=mode)
            else:
                with self.assertRaisesRegex(
                    ValueError, "Variable 'y' derivative == 0 in constraint 'c'"
                ):
                    calculate_variable_from_constraint(m.y, m.c, diff_mode=mode)

        # should succeed with or without a linesearch
        m.e = Constraint(expr=(m.x - 2.0) ** 2 - 1 == 0)
        for mode in all_diff_modes:
            m.x.set_value(3.1)
            calculate_variable_from_constraint(
                m.x, m.e, linesearch=False, diff_mode=mode
            )
            self.assertAlmostEqual(value(m.x), 3)

        for mode in all_diff_modes:
            m.x.set_value(3.1)
            calculate_variable_from_constraint(
                m.x, m.e, linesearch=True, diff_mode=mode
            )
            self.assertAlmostEqual(value(m.x), 3)

        # we expect this to succeed with the linesearch
        m.f = Constraint(expr=1.0 / (1.0 + exp(-m.x)) - 0.5 == 0)
        for mode in all_diff_modes:
            m.x.set_value(3.0)
            calculate_variable_from_constraint(
                m.x, m.f, linesearch=True, diff_mode=mode
            )
            self.assertAlmostEqual(value(m.x), 0)

        # we expect this to fail without a linesearch
        for mode in all_diff_modes:
            m.x.set_value(3.0)
            with self.assertRaisesRegex(
                RuntimeError,
                "Newton's method encountered a derivative of constraint 'f' "
                "with respect to variable 'x' that was too close to zero",
            ):
                calculate_variable_from_constraint(
                    m.x, m.f, linesearch=False, diff_mode=mode
                )

        # Calculate the bubble point of Benzene.  THe first step
        # computed by calculate_variable_from_constraint will make the
        # second term become complex, and the evaluation will fail.
        # This tests that the algorithm cleanly continues
        m = ConcreteModel()
        m.x = Var()
        m.pc = 48.9e5
        m.tc = 562.2
        m.psc = {'A': -6.98273, 'B': 1.33213, 'C': -2.62863, 'D': -3.33399}
        m.p = 101325

        @m.Constraint()
        def f(m):
            return (
                m.pc
                * exp(
                    (
                        m.psc['A'] * (1 - m.x / m.tc)
                        + m.psc['B'] * (1 - m.x / m.tc) ** 1.5
                        + m.psc['C'] * (1 - m.x / m.tc) ** 3
                        + m.psc['D'] * (1 - m.x / m.tc) ** 6
                    )
                    / (1 - (1 - m.x / m.tc))
                )
                - m.p
                == 0
            )

        m.x.set_value(298.15)
        calculate_variable_from_constraint(m.x, m.f, linesearch=False)
        self.assertAlmostEqual(value(m.x), 353.31855602)
        m.x.set_value(298.15)
        calculate_variable_from_constraint(m.x, m.f, linesearch=True)
        self.assertAlmostEqual(value(m.x), 353.31855602)

        # Starting with an invalid guess (above TC) should raise an
        # exception
        m.x.set_value(600)
        output = StringIO()
        with LoggingIntercept(output, 'pyomo', logging.WARNING):
            with self.assertRaises(TypeError):
                calculate_variable_from_constraint(m.x, m.f, linesearch=False)
        self.assertIn(
            'Encountered an error evaluating the expression at the initial guess',
            output.getvalue(),
        )

        # This example triggers an expression evaluation error if the
        # linesearch is turned off because the first step in Newton's
        # method will cause the LHS to become complex
        m = ConcreteModel()
        m.x = Var()
        m.c = Constraint(expr=(1 / m.x**3) ** 0.5 == 100)
        m.x = 0.1
        calculate_variable_from_constraint(m.x, m.c, linesearch=True)
        self.assertAlmostEqual(value(m.x), 0.046415888)
        m.x = 0.1
        output = StringIO()
        with LoggingIntercept(output, 'pyomo', logging.WARNING):
            with self.assertRaises(ValueError):
                # Note that the ValueError is different between Python 2
                # and Python 3: in Python 2 it is a specific error
                # "negative number cannot be raised to a fractional
                # power", and We mock up that error in Python 3 by
                # raising a generic ValueError in
                # calculate_variable_from_constraint
                calculate_variable_from_constraint(m.x, m.c, linesearch=False)
        self.assertIn(
            "Newton's method encountered an error evaluating the expression for "
            "constraint 'c'.",
            output.getvalue(),
        )

        # This is a completely contrived example where the linesearch
        # hits the iteration limit before Newton's method ever finds a
        # feasible step
        m = ConcreteModel()
        m.x = Var()
        m.c = Constraint(expr=m.x**0.5 == -1e-8)
        m.x = 1e-8  # 197.932807183
        with self.assertRaisesRegex(
            IterationLimitError,
            "Linesearch iteration limit reached solving for variable 'x' using "
            "constraint 'c'; remaining residual = {function evaluation error}",
        ):
            calculate_variable_from_constraint(m.x, m.c, linesearch=True, alpha_min=0.5)

    def test_bound_violation(self):
        # Test Issue #2176: solving a constraint where the intermediate
        # value can step outside the bounds
        m = ConcreteModel()
        m.v1 = Var(initialize=1, domain=NonNegativeReals)
        m.c1 = Constraint(expr=m.v1 == 0)

        # Calculate value of v1 using constraint c1
        for mode in all_diff_modes:
            m.v1.set_value(None)
            calculate_variable_from_constraint(m.v1, m.c1, diff_mode=mode)
            self.assertEqual(value(m.v1), 0)

        # Calculate value of v1 using a scaled constraint c2
        m.c2 = Constraint(expr=m.v1 * 10 == 0)
        for mode in all_diff_modes:
            m.v1.set_value(1)
            calculate_variable_from_constraint(m.v1, m.c2, diff_mode=mode)
            self.assertEqual(value(m.v1), 0)

        # Test linear solution falling outside bounds
        m.c3 = Constraint(expr=m.v1 * 10 == -1)
        for mode in all_diff_modes:
            m.v1.set_value(1)
            calculate_variable_from_constraint(m.v1, m.c3, diff_mode=mode)
            self.assertEqual(value(m.v1), -0.1)

    @unittest.skipUnless(differentiate_available, "this test requires sympy")
    def test_nonlinear_bound_violation(self):
        m = ConcreteModel()
        m.v1 = Var(initialize=1, domain=NonNegativeReals)
        m.c1 = Constraint(expr=m.v1 == 0)

        # Test nonlinear solution falling outside bounds
        m.c4 = Constraint(expr=m.v1**3 == -8)
        for mode in all_diff_modes:
            m.v1.set_value(1)
            calculate_variable_from_constraint(m.v1, m.c4, diff_mode=mode)
            self.assertEqual(value(m.v1), -2)

    def test_warn_final_value_linear(self):
        m = ConcreteModel()
        m.x = Var(bounds=(0, 1))
        m.c1 = Constraint(expr=m.x == 10)
        m.c2 = Constraint(expr=5 * m.x == 10)

        with LoggingIntercept() as LOG:
            calculate_variable_from_constraint(m.x, m.c1)
        self.assertEqual(
            LOG.getvalue().strip(),
            "Setting Var 'x' to a numeric value `10` outside the bounds (0, 1).",
        )
        self.assertEqual(value(m.x), 10)

        with LoggingIntercept() as LOG:
            calculate_variable_from_constraint(m.x, m.c2)
        self.assertEqual(
            LOG.getvalue().strip(),
            "Setting Var 'x' to a numeric value `2.0` outside the bounds (0, 1).",
        )
        self.assertEqual(value(m.x), 2)

    @unittest.skipUnless(differentiate_available, "this test requires sympy")
    def test_warn_final_value_nonlinear(self):
        m = ConcreteModel()
        m.x = Var(bounds=(0, 1))
        m.c3 = Constraint(expr=(m.x - 3.5) ** 2 == 0)

        with LoggingIntercept() as LOG:
            calculate_variable_from_constraint(m.x, m.c3)
        self.assertRegex(
            LOG.getvalue().strip(),
            r"Setting Var 'x' to a numeric value `[0-9\.]+` outside the "
            r"bounds \(0, 1\).",
        )
        self.assertAlmostEqual(value(m.x), 3.5, 3)

        m.x.domain = Binary
        with LoggingIntercept() as LOG:
            calculate_variable_from_constraint(m.x, m.c3)
        self.assertRegex(
            LOG.getvalue().strip(),
            r"Setting Var 'x' to a value `[0-9\.]+` \(float\) not in domain Binary.",
        )
        self.assertAlmostEqual(value(m.x), 3.5, 3)

    @unittest.skipUnless(differentiate_available, "this test requires sympy")
    def test_nonlinear_overflow(self):
        # Regression check to make sure calculate_variable_from_constraint
        # can handle extreme non-linear cases where assuming linear behaviour
        # results in OverflowErrors
        m = ConcreteModel()
        m.x = Var(initialize=1)
        m.c = Constraint(expr=exp(1e2 * m.x**2) == 100)

        calculate_variable_from_constraint(m.x, m.c)

        self.assertAlmostEqual(value(m.x), 0.214597, 5)

    def test_external_function(self):
        m = ConcreteModel()
        m.x = Var()
        m.sq = ExternalFunction(fgh=sum_sq)
        m.c = Constraint(expr=m.sq(m.x - 3) == 0)

        with LoggingIntercept(level=logging.DEBUG) as LOG:
            calculate_variable_from_constraint(m.x, m.c)
        self.assertAlmostEqual(value(m.x), 3, 3)
        self.assertEqual(
            LOG.getvalue(),
            "Calculating symbolic derivative of expression failed. "
            "Reverting to numeric differentiation\n",
        )

    @unittest.skipUnless(sympy_available, 'test expects that sympy is available')
    def test_external_function_explicit_sympy(self):
        m = ConcreteModel()
        m.x = Var()
        m.sq = ExternalFunction(fgh=sum_sq)
        m.c = Constraint(expr=m.sq(m.x - 3) == 0)

        with self.assertRaisesRegex(
            TypeError,
            r"Expressions containing external functions are not convertible "
            r"to sympy expressions \(found 'f\(x0 - 3",
        ):
            calculate_variable_from_constraint(
                m.x, m.c, diff_mode=differentiate.Modes.sympy
            )

    def test_linea_search_overflow(self):
        # This tests the example from #3540
        m = ConcreteModel()
        m.x = Var(initialize=5)
        m.y = Var(initialize=1)
        m.con = Constraint(expr=m.y == 10 ** (-m.x))

        calculate_variable_from_constraint(m.x, m.con)
        self.assertAlmostEqual(value(m.x), 0)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os
import pyomo.common.unittest as unittest
from io import StringIO
import logging

import pyomo.environ as pyo
from pyomo.common.dependencies import (
    numpy as np,
    numpy_available,
    pandas as pd,
    pandas_available,
)
from pyomo.contrib.sensitivity_toolbox.sens import SensitivityInterface
from pyomo.contrib.sensitivity_toolbox.k_aug import K_augInterface

opt_ipopt = pyo.SolverFactory('ipopt', solver_io='nl', validate=False)
opt_k_aug = pyo.SolverFactory('k_aug', solver_io='nl', validate=False)
opt_dot_sens = pyo.SolverFactory('dot_sens', solver_io='nl', validate=False)


def simple_model_1():
    m = pyo.ConcreteModel()
    m.v1 = pyo.Var(initialize=10.0)
    m.v2 = pyo.Var(initialize=10.0)

    m.p = pyo.Param(mutable=True, initialize=1.0)

    m.eq_con = pyo.Constraint(expr=m.v1 * m.v2 - m.p == 0)

    m.obj = pyo.Objective(expr=m.v1**2 + m.v2**2, sense=pyo.minimize)

    return m


class TestK_augInterface(unittest.TestCase):
    @unittest.skipIf(not opt_k_aug.available(), "k_aug is not available")
    def test_clear_dir_k_aug(self):
        m = simple_model_1()
        sens = SensitivityInterface(m, clone_model=False)
        k_aug = K_augInterface()

        opt_ipopt.solve(m, tee=True)
        m.ptb = pyo.Param(mutable=True, initialize=1.5)

        cwd = os.getcwd()
        dir_contents = os.listdir(cwd)

        sens_param = [m.p]
        sens.setup_sensitivity(sens_param)

        k_aug.k_aug(m, tee=True)

        # We are back in our working directory
        self.assertEqual(cwd, os.getcwd())

        # The contents of this directory have not changed
        self.assertEqual(dir_contents, os.listdir(cwd))

        # In particular, the following files do not exist
        self.assertFalse(os.path.exists("dsdp_in_.in"))
        self.assertFalse(os.path.exists("conorder.txt"))
        self.assertFalse(os.path.exists("timings_k_aug_dsdp.txt"))

        # But they have been transferred to our k_aug interface's data
        # dict as strings.
        self.assertIsInstance(k_aug.data["dsdp_in_.in"], str)
        self.assertIsInstance(k_aug.data["conorder.txt"], str)
        self.assertIsInstance(k_aug.data["timings_k_aug_dsdp.txt"], str)

    @unittest.skipIf(not opt_k_aug.available(), "k_aug is not available")
    @unittest.skipIf(not opt_dot_sens.available(), "dot_sens is not available")
    def test_clear_dir_dot_sens(self):
        m = simple_model_1()
        sens = SensitivityInterface(m, clone_model=False)
        k_aug = K_augInterface()
        opt_ipopt.solve(m, tee=True)
        m.ptb = pyo.Param(mutable=True, initialize=1.5)

        cwd = os.getcwd()
        dir_contents = os.listdir(cwd)

        sens_param = [m.p]
        sens.setup_sensitivity(sens_param)

        # Call k_aug
        k_aug.k_aug(m, tee=True)
        self.assertIsInstance(k_aug.data["dsdp_in_.in"], str)

        sens.perturb_parameters([m.ptb])

        # Call dot_sens. In the process, we re-write dsdp_in_.in
        k_aug.dot_sens(m, tee=True)

        # Make sure we get the values we expect. This problem is easy enough
        # to solve by hand:
        # x = [1, 1, -2] = [v1, v2, dual]
        # Sensitivity system:
        # | 2 -2  1 |
        # |-2  2  1 | dx/dp = -[dL/dxdp, dc/dp]^T = -[0, 0, -1]^T
        # | 1  1  0 |
        # => dx/dp = [0.5, 0.5, 0]^T
        # Here, dp = [0.5]
        # => dx = [0.25, 0.25, 0]^T
        # => x_new = [1.25, 1.25, -2]
        self.assertAlmostEqual(m.v1.value, 1.25, 7)
        self.assertAlmostEqual(m.v2.value, 1.25, 7)

        # We are back in our working directory
        self.assertEqual(cwd, os.getcwd())

        # The contents of this directory have not changed
        self.assertEqual(dir_contents, os.listdir(cwd))
        self.assertFalse(os.path.exists("dsdp_in_.in"))
        self.assertFalse(os.path.exists("delta_p.out"))
        self.assertFalse(os.path.exists("dot_out.out"))
        self.assertFalse(os.path.exists("timings_dot_driver_dsdp.txt"))

        # And we have saved strings of the file contents.
        self.assertIsInstance(k_aug.data["dsdp_in_.in"], str)
        self.assertIsInstance(k_aug.data["delta_p.out"], str)
        self.assertIsInstance(k_aug.data["dot_out.out"], str)
        self.assertIsInstance(k_aug.data["timings_dot_driver_dsdp.txt"], str)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.fileutils import find_library
from pyomo.common.dependencies import numpy as np
from pyomo.contrib.pynumero.exceptions import PyNumeroEvaluationError
import ctypes
import logging
import os

logger = logging.getLogger(__name__)

CURRENT_INTERFACE_VERSION = 3


class _NotSet:
    pass


def _LoadASLInterface(libname):
    ASLib = ctypes.cdll.LoadLibrary(libname)

    # define 1d array
    array_1d_double = np.ctypeslib.ndpointer(
        dtype=np.double, ndim=1, flags='CONTIGUOUS'
    )
    array_1d_int = np.ctypeslib.ndpointer(dtype=np.intc, ndim=1, flags='CONTIGUOUS')

    # library version
    try:
        ASLib.EXTERNAL_AmplInterface_version.argtypes = None
        ASLib.EXTERNAL_AmplInterface_version.restype = ctypes.c_int
        interface_version = ASLib.EXTERNAL_AmplInterface_version()
    except AttributeError:
        interface_version = 1

    # ASL version
    if interface_version >= 3:
        ASLib.EXTERNAL_get_asl_date.argtypes = []
        ASLib.EXTERNAL_get_asl_date.restype = ctypes.c_long

    # constructor
    ASLib.EXTERNAL_AmplInterface_new.argtypes = [ctypes.c_char_p]
    ASLib.EXTERNAL_AmplInterface_new.restype = ctypes.c_void_p

    if interface_version >= 2:
        ASLib.EXTERNAL_AmplInterface_new_file.argtypes = [
            ctypes.c_char_p,
            ctypes.c_char_p,
        ]
    else:
        ASLib.EXTERNAL_AmplInterface_new_file.argtypes = [ctypes.c_char_p]
    ASLib.EXTERNAL_AmplInterface_new_file.restype = ctypes.c_void_p

    # ASLib.EXTERNAL_AmplInterface_new_str.argtypes = [ctypes.c_char_p]
    # ASLib.EXTERNAL_AmplInterface_new_str.restype = ctypes.c_void_p

    # number of variables
    ASLib.EXTERNAL_AmplInterface_n_vars.argtypes = [ctypes.c_void_p]
    ASLib.EXTERNAL_AmplInterface_n_vars.restype = ctypes.c_int

    # number of constraints
    ASLib.EXTERNAL_AmplInterface_n_constraints.argtypes = [ctypes.c_void_p]
    ASLib.EXTERNAL_AmplInterface_n_constraints.restype = ctypes.c_int

    # number of nonzeros in jacobian
    ASLib.EXTERNAL_AmplInterface_nnz_jac_g.argtypes = [ctypes.c_void_p]
    ASLib.EXTERNAL_AmplInterface_nnz_jac_g.restype = ctypes.c_int

    # number of nonzeros in hessian of lagrangian
    ASLib.EXTERNAL_AmplInterface_nnz_hessian_lag.argtypes = [ctypes.c_void_p]
    ASLib.EXTERNAL_AmplInterface_nnz_hessian_lag.restype = ctypes.c_int

    # lower bounds on x
    ASLib.EXTERNAL_AmplInterface_x_lower_bounds.argtypes = [
        ctypes.c_void_p,
        array_1d_double,
        ctypes.c_int,
    ]
    ASLib.EXTERNAL_AmplInterface_x_lower_bounds.restype = None

    # upper bounds on x
    ASLib.EXTERNAL_AmplInterface_x_upper_bounds.argtypes = [
        ctypes.c_void_p,
        array_1d_double,
        ctypes.c_int,
    ]
    ASLib.EXTERNAL_AmplInterface_x_upper_bounds.restype = None

    # lower bounds on g
    ASLib.EXTERNAL_AmplInterface_g_lower_bounds.argtypes = [
        ctypes.c_void_p,
        array_1d_double,
        ctypes.c_int,
    ]
    ASLib.EXTERNAL_AmplInterface_g_lower_bounds.restype = None

    # upper bounds on g
    ASLib.EXTERNAL_AmplInterface_g_upper_bounds.argtypes = [
        ctypes.c_void_p,
        array_1d_double,
        ctypes.c_int,
    ]
    ASLib.EXTERNAL_AmplInterface_g_upper_bounds.restype = None

    # initial value x
    ASLib.EXTERNAL_AmplInterface_get_init_x.argtypes = [
        ctypes.c_void_p,
        array_1d_double,
        ctypes.c_int,
    ]
    ASLib.EXTERNAL_AmplInterface_get_init_x.restype = None

    # initial value multipliers
    ASLib.EXTERNAL_AmplInterface_get_init_multipliers.argtypes = [
        ctypes.c_void_p,
        array_1d_double,
        ctypes.c_int,
    ]
    ASLib.EXTERNAL_AmplInterface_get_init_multipliers.restype = None

    # evaluate objective
    ASLib.EXTERNAL_AmplInterface_eval_f.argtypes = [
        ctypes.c_void_p,
        array_1d_double,
        ctypes.c_int,
        ctypes.POINTER(ctypes.c_double),
    ]
    ASLib.EXTERNAL_AmplInterface_eval_f.restype = ctypes.c_bool

    # gradient objective
    ASLib.EXTERNAL_AmplInterface_eval_deriv_f.argtypes = [
        ctypes.c_void_p,
        array_1d_double,
        array_1d_double,
        ctypes.c_int,
    ]
    ASLib.EXTERNAL_AmplInterface_eval_deriv_f.restype = ctypes.c_bool

    # structure jacobian of constraints
    ASLib.EXTERNAL_AmplInterface_struct_jac_g.argtypes = [
        ctypes.c_void_p,
        array_1d_int,
        array_1d_int,
        ctypes.c_int,
    ]
    ASLib.EXTERNAL_AmplInterface_struct_jac_g.restype = None

    # structure hessian of Lagrangian
    ASLib.EXTERNAL_AmplInterface_struct_hes_lag.argtypes = [
        ctypes.c_void_p,
        array_1d_int,
        array_1d_int,
        ctypes.c_int,
    ]
    ASLib.EXTERNAL_AmplInterface_struct_hes_lag.restype = None

    # evaluate constraints
    ASLib.EXTERNAL_AmplInterface_eval_g.argtypes = [
        ctypes.c_void_p,
        array_1d_double,
        ctypes.c_int,
        array_1d_double,
        ctypes.c_int,
    ]
    ASLib.EXTERNAL_AmplInterface_eval_g.restype = ctypes.c_bool

    # evaluate jacobian constraints
    ASLib.EXTERNAL_AmplInterface_eval_jac_g.argtypes = [
        ctypes.c_void_p,
        array_1d_double,
        ctypes.c_int,
        array_1d_double,
        ctypes.c_int,
    ]
    ASLib.EXTERNAL_AmplInterface_eval_jac_g.restype = ctypes.c_bool

    # temporary try/except block while changes get merged in pynumero_libraries
    try:
        ASLib.EXTERNAL_AmplInterface_dummy.argtypes = [ctypes.c_void_p]
        ASLib.EXTERNAL_AmplInterface_dummy.restype = None
        # evaluate hessian Lagrangian
        ASLib.EXTERNAL_AmplInterface_eval_hes_lag.argtypes = [
            ctypes.c_void_p,
            array_1d_double,
            ctypes.c_int,
            array_1d_double,
            ctypes.c_int,
            array_1d_double,
            ctypes.c_int,
            ctypes.c_double,
        ]
        ASLib.EXTERNAL_AmplInterface_eval_hes_lag.restype = ctypes.c_bool
    except Exception:
        # evaluate hessian Lagrangian
        ASLib.EXTERNAL_AmplInterface_eval_hes_lag.argtypes = [
            ctypes.c_void_p,
            array_1d_double,
            ctypes.c_int,
            array_1d_double,
            ctypes.c_int,
            array_1d_double,
            ctypes.c_int,
        ]
        ASLib.EXTERNAL_AmplInterface_eval_hes_lag.restype = ctypes.c_bool
        interface_version = 0

    # finalize solution
    ASLib.EXTERNAL_AmplInterface_finalize_solution.argtypes = [
        ctypes.c_void_p,
        ctypes.c_int,
        ctypes.c_char_p,
        array_1d_double,
        ctypes.c_int,
        array_1d_double,
        ctypes.c_int,
    ]
    ASLib.EXTERNAL_AmplInterface_finalize_solution.restype = None

    # destructor
    ASLib.EXTERNAL_AmplInterface_free_memory.argtypes = [ctypes.c_void_p]
    ASLib.EXTERNAL_AmplInterface_free_memory.restype = None

    if CURRENT_INTERFACE_VERSION != interface_version:
        logger.warning(
            'The current pynumero_ASL library is version=%s, but found '
            'version=%s.  Please recompile / update your pynumero_ASL '
            'library.' % (CURRENT_INTERFACE_VERSION, interface_version)
        )

    return ASLib, interface_version


class AmplInterface:
    libname = _NotSet
    ASLib = None
    interface_version = None
    asl_date = None

    @classmethod
    def available(cls):
        if cls.libname is _NotSet:
            cls.libname = find_library('pynumero_ASL')
        if cls.libname is None:
            return False
        return os.path.exists(cls.libname)

    def __init__(self, filename=None, nl_buffer=None):
        if not AmplInterface.available():
            raise RuntimeError("Cannot load the PyNumero ASL interface (pynumero_ASL)")

        if nl_buffer is not None:
            raise NotImplementedError(
                "AmplInterface only supported form NL-file for now"
            )

        # Be sure to remove AMPLFUNC from the environment before loading
        # the ASL.  This should prevent it from potentially caching an
        # AMPLFUNC from the initial load and letting it bleed into
        # (potentially unrelated) subsequent instances
        amplfunc = os.environ.pop('AMPLFUNC', '')

        if AmplInterface.ASLib is None:
            AmplInterface.ASLib, AmplInterface.interface_version = _LoadASLInterface(
                AmplInterface.libname
            )
            if AmplInterface.interface_version >= 3:
                AmplInterface.asl_date = AmplInterface.ASLib.EXTERNAL_get_asl_date()
            else:
                AmplInterface.asl_date = 0

        if filename is not None:
            if nl_buffer is not None:
                raise ValueError("Cannot specify both filename= and nl_buffer=")

            b_data = filename.encode('utf-8')
            if self.interface_version >= 2:
                args = (b_data, amplfunc.encode('utf-8'))
            else:
                # Old ASL interface library.
                if amplfunc:
                    # we need to put AMPLFUNC back into the environment,
                    # as old versions of the library rely on ONLY the
                    # environment variable for passing the library(ies)
                    # locations to the ASL
                    os.environ['AMPLFUNC'] = amplfunc
                args = (b_data,)
            self._obj = self.ASLib.EXTERNAL_AmplInterface_new_file(*args)
        elif nl_buffer is not None:
            b_data = nl_buffer.encode('utf-8')
            if os.name in ['nt', 'dos']:
                self._obj = self.ASLib.EXTERNAL_AmplInterface_new_file(b_data)
            else:
                self._obj = self.ASLib.EXTERNAL_AmplInterface_new_str(b_data)

        assert self._obj, "Error building ASL interface. Possible error in nl-file"

        self._nx = self.get_n_vars()
        self._ny = self.get_n_constraints()
        self._nnz_jac_g = self.get_nnz_jac_g()
        self._nnz_hess = self.get_nnz_hessian_lag()

    def __del__(self):
        self.ASLib.EXTERNAL_AmplInterface_free_memory(self._obj)

    def get_n_vars(self):
        return self.ASLib.EXTERNAL_AmplInterface_n_vars(self._obj)

    def get_n_constraints(self):
        return self.ASLib.EXTERNAL_AmplInterface_n_constraints(self._obj)

    def get_nnz_jac_g(self):
        return self.ASLib.EXTERNAL_AmplInterface_nnz_jac_g(self._obj)

    def get_nnz_hessian_lag(self):
        return self.ASLib.EXTERNAL_AmplInterface_nnz_hessian_lag(self._obj)

    def get_bounds_info(self, xl, xu, gl, gu):
        x_l = xl.astype(np.double, casting='safe', copy=False)
        x_u = xu.astype(np.double, casting='safe', copy=False)
        g_l = gl.astype(np.double, casting='safe', copy=False)
        g_u = gu.astype(np.double, casting='safe', copy=False)
        nx = len(x_l)
        ng = len(g_l)
        assert nx == len(x_u), "lower and upper bound x vectors must be the same size"
        assert ng == len(g_u), "lower and upper bound g vectors must be the same size"
        self.ASLib.EXTERNAL_AmplInterface_get_bounds_info(
            self._obj, x_l, x_u, nx, g_l, g_u, ng
        )

    def get_x_lower_bounds(self, invec):
        self.ASLib.EXTERNAL_AmplInterface_x_lower_bounds(self._obj, invec, len(invec))

    def get_x_upper_bounds(self, invec):
        self.ASLib.EXTERNAL_AmplInterface_x_upper_bounds(self._obj, invec, len(invec))

    def get_g_lower_bounds(self, invec):
        self.ASLib.EXTERNAL_AmplInterface_g_lower_bounds(self._obj, invec, len(invec))

    def get_g_upper_bounds(self, invec):
        self.ASLib.EXTERNAL_AmplInterface_g_upper_bounds(self._obj, invec, len(invec))

    def get_init_x(self, invec):
        self.ASLib.EXTERNAL_AmplInterface_get_init_x(self._obj, invec, len(invec))

    def get_init_multipliers(self, invec):
        self.ASLib.EXTERNAL_AmplInterface_get_init_multipliers(
            self._obj, invec, len(invec)
        )

    def eval_f(self, x):
        assert x.size == self._nx, "Error: Dimension mismatch."
        assert (
            x.dtype == np.double
        ), "Error: array type. Function eval_deriv_f expects an array of type double"
        sol = ctypes.c_double()
        res = self.ASLib.EXTERNAL_AmplInterface_eval_f(
            self._obj, x, self._nx, ctypes.byref(sol)
        )
        if not res:
            raise PyNumeroEvaluationError("Error in AMPL evaluation")
        return sol.value

    def eval_deriv_f(self, x, df):
        assert x.size == self._nx, "Error: Dimension mismatch."
        assert (
            x.dtype == np.double
        ), "Error: array type. Function eval_deriv_f expects an array of type double"
        res = self.ASLib.EXTERNAL_AmplInterface_eval_deriv_f(self._obj, x, df, len(x))
        if not res:
            raise PyNumeroEvaluationError("Error in AMPL evaluation")

    def struct_jac_g(self, irow, jcol):
        irow_p = irow.astype(np.intc, casting='safe', copy=False)
        jcol_p = jcol.astype(np.intc, casting='safe', copy=False)
        assert len(irow) == len(
            jcol
        ), "Error: Dimension mismatch. Arrays irow and jcol must be of the same size"
        assert (
            len(irow) == self._nnz_jac_g
        ), "Error: Dimension mismatch. Jacobian has {} nnz".format(self._nnz_jac_g)
        self.ASLib.EXTERNAL_AmplInterface_struct_jac_g(
            self._obj, irow_p, jcol_p, len(irow)
        )

    def struct_hes_lag(self, irow, jcol):
        irow_p = irow.astype(np.intc, casting='safe', copy=False)
        jcol_p = jcol.astype(np.intc, casting='safe', copy=False)
        assert len(irow) == len(
            jcol
        ), "Error: Dimension mismatch. Arrays irow and jcol must be of the same size"
        assert (
            len(irow) == self._nnz_hess
        ), "Error: Dimension mismatch. Hessian has {} nnz".format(self._nnz_hess)
        self.ASLib.EXTERNAL_AmplInterface_struct_hes_lag(
            self._obj, irow_p, jcol_p, len(irow)
        )

    def eval_jac_g(self, x, jac_g_values):
        assert x.size == self._nx, "Error: Dimension mismatch."
        assert jac_g_values.size == self._nnz_jac_g, "Error: Dimension mismatch."
        xeval = x.astype(np.double, casting='safe', copy=False)
        jac_eval = jac_g_values.astype(np.double, casting='safe', copy=False)
        res = self.ASLib.EXTERNAL_AmplInterface_eval_jac_g(
            self._obj, xeval, self._nx, jac_eval, self._nnz_jac_g
        )
        if not res:
            raise PyNumeroEvaluationError("Error in AMPL evaluation")

    def eval_g(self, x, g):
        assert x.size == self._nx, "Error: Dimension mismatch."
        assert g.size == self._ny, "Error: Dimension mismatch."
        assert (
            x.dtype == np.double
        ), "Error: array type. Function eval_g expects an array of type double"
        assert (
            g.dtype == np.double
        ), "Error: array type. Function eval_g expects an array of type double"
        res = self.ASLib.EXTERNAL_AmplInterface_eval_g(
            self._obj, x, self._nx, g, self._ny
        )
        if not res:
            raise PyNumeroEvaluationError("Error in AMPL evaluation")

    def eval_hes_lag(self, x, lam, hes_lag, obj_factor=1.0):
        assert x.size == self._nx, "Error: Dimension mismatch."
        assert lam.size == self._ny, "Error: Dimension mismatch."
        assert hes_lag.size == self._nnz_hess, "Error: Dimension mismatch."
        assert (
            x.dtype == np.double
        ), "Error: array type. Function eval_hes_lag expects an array of type double"
        assert (
            lam.dtype == np.double
        ), "Error: array type. Function eval_hes_lag expects an array of type double"
        assert (
            hes_lag.dtype == np.double
        ), "Error: array type. Function eval_hes_lag expects an array of type double"
        if self.interface_version >= 1:
            res = self.ASLib.EXTERNAL_AmplInterface_eval_hes_lag(
                self._obj,
                x,
                self._nx,
                lam,
                self._ny,
                hes_lag,
                self._nnz_hess,
                obj_factor,
            )
        else:
            res = self.ASLib.EXTERNAL_AmplInterface_eval_hes_lag(
                self._obj, x, self._nx, lam, self._ny, hes_lag, self._nnz_hess
            )
        if not res:
            raise PyNumeroEvaluationError("Error in AMPL evaluation")

    def finalize_solution(self, ampl_solve_status_num, msg, x, lam):
        b_msg = msg.encode('utf-8')
        self.ASLib.EXTERNAL_AmplInterface_finalize_solution(
            self._obj, ampl_solve_status_num, b_msg, x, len(x), lam, len(lam)
        )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

_init_url = "$URL$"

# NOTE: releaselevel should be left at 'invalid' for trunk development
#     and set to 'final' for releases.  During development, the
#     major.minor.micro should point to the NEXT release (generally, the
#     next micro release after the current release).
#
# Note: When cutting a release, also update the major/minor/micro in
#
#     pyomo/RELEASE.md
#
# The VOTD zipbuilder will automatically change releaselevel to "VOTD
# {hash}" and set the serial number to YYMMDDhhmm.  The serial number
# should generally be left at 0, unless a downstream package is tracking
# main and needs a hard reference to "suitably new" development.
major = 6
minor = 10
micro = 0
# releaselevel = 'invalid'
releaselevel = 'final'
serial = 0

if releaselevel == 'final':
    pass
elif '/tags/' in _init_url:  # pragma:nocover
    releaselevel = 'final'
elif releaselevel == 'invalid':
    from os.path import exists as _exists, join as _join, dirname as _dirname

    if __file__.endswith('setup.py'):
        # This file is being sourced (exec'ed) from setup.py.
        # dirname(__file__) setup.py's scope is the root source directory
        _rootdir = _dirname(__file__)
    else:
        # Ideally, this should import PYOMO_ROOT_DIR from pyomo.common
        # instead of reimplementing that logic here.  Unfortunately,
        # there is a circular reference (pyomo.common.log imports
        # releaselevel).  We will leave this module completely
        # independent of the rest of Pyomo.
        #
        # __file__ fails if script is called in different ways on Windows
        # __file__ fails if someone does os.chdir() before
        # sys.argv[0] also fails because it doesn't always contain the path
        from inspect import getfile as _getfile, currentframe as _frame
        from os.path import abspath as _abspath

        _rootdir = _join(_dirname(_abspath(_getfile(_frame()))), '..', '..')

    if _exists(_join(_rootdir, '.git')):
        try:
            with open(_join(_rootdir, '.git', 'HEAD')) as _FILE:
                _ref = _FILE.readline().strip()
            releaselevel = 'devel {%s}' % (_ref.split('/')[-1].split('\\')[-1],)
        except:
            releaselevel = 'devel'
    else:
        releaselevel = 'VOTD'


version_info = (major, minor, micro, releaselevel, serial)

__version__ = '.'.join(str(x) for x in version_info[:3])
if releaselevel.startswith('devel'):
    __version__ += ".dev%d" % (serial,)
elif releaselevel.startswith('VOTD'):
    __version__ += "a%d" % (serial,)

version = __version__
if releaselevel != 'final':
    version += ' (' + releaselevel + ')'



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.kernel as pmo
from pyomo.core import (
    ConcreteModel,
    Var,
    Objective,
    Constraint,
    NonNegativeReals,
    maximize,
    ConstraintList,
)
from pyomo.opt import TerminationCondition
from pyomo.solvers.tests.models.base import _BaseTestModel, register_model


@register_model
class QCP_simple(_BaseTestModel):
    """
    A continuous model with a quadratic objective and quadratics constraints
    """

    description = "QCP_simple"
    level = ('nightly', 'expensive')
    capabilities = set(['linear', 'quadratic_objective', 'quadratic_constraint'])

    def __init__(self):
        _BaseTestModel.__init__(self)
        self.add_results(self.description + ".json")

    def _generate_model(self):
        self.model = ConcreteModel()
        model = self.model
        model._name = self.description

        model.x = Var(within=NonNegativeReals)
        model.y = Var(within=NonNegativeReals)
        model.z = Var(within=NonNegativeReals)
        model.fixed_var = Var()
        model.fixed_var.fix(0.2)
        model.q1 = Var(bounds=(None, 0.2))
        model.q2 = Var(bounds=(-2, None))
        model.obj = Objective(expr=model.x + model.q1 - model.q2, sense=maximize)
        model.c0 = Constraint(expr=model.x + model.y + model.z == 1)
        model.qc0 = Constraint(
            expr=model.x**2 + model.y**2 + model.fixed_var <= model.z**2
        )
        model.qc1 = Constraint(expr=model.x**2 <= model.y * model.z)
        model.c = ConstraintList()
        model.c.add((0, -model.q1**2 + model.fixed_var, None))
        model.c.add((None, model.q2**2 + model.fixed_var, 5))

    def warmstart_model(self):
        assert self.model is not None
        model = self.model
        model.x.value = 1
        model.y.value = 1
        model.z.value = 1

    def post_solve_test_validation(self, tester, results):
        if tester is None:
            assert results['Solver'][0]['termination condition'] in (
                TerminationCondition.optimal,
                TerminationCondition.locallyOptimal,
            )
        else:
            tester.assertIn(
                results['Solver'][0]['termination condition'],
                (TerminationCondition.optimal, TerminationCondition.locallyOptimal),
            )


@register_model
class QCP_simple_nosuffixes(QCP_simple):
    description = "QCP_simple_nosuffixes"
    test_pickling = False

    def __init__(self):
        QCP_simple.__init__(self)
        self.disable_suffix_tests = True
        self.add_results("QCP_simple.json")


@register_model
class QCP_simple_kernel(QCP_simple):
    def _generate_model(self):
        self.model = pmo.block()
        model = self.model
        model._name = self.description

        model.x = pmo.variable(domain=NonNegativeReals)
        model.y = pmo.variable(domain=NonNegativeReals)
        model.z = pmo.variable(domain=NonNegativeReals)
        model.fixed_var = pmo.variable()
        model.fixed_var.fix(0.2)
        model.q1 = pmo.variable(ub=0.2)
        model.q2 = pmo.variable(lb=-2)
        model.obj = pmo.objective(model.x + model.q1 - model.q2, sense=maximize)
        model.c0 = pmo.constraint(model.x + model.y + model.z == 1)
        model.qc0 = pmo.constraint(
            model.x**2 + model.y**2 + model.fixed_var <= model.z**2
        )
        model.qc1 = pmo.constraint(model.x**2 <= model.y * model.z)
        model.c = pmo.constraint_dict()
        model.c[1] = pmo.constraint(lb=0, body=-model.q1**2 + model.fixed_var)
        model.c[2] = pmo.constraint(body=model.q2**2 + model.fixed_var, ub=5)


@register_model
class QCP_simple_nosuffixes_kernel(QCP_simple_kernel):
    description = "QCP_simple_nosuffixes"
    test_pickling = False

    def __init__(self):
        QCP_simple.__init__(self)
        self.disable_suffix_tests = True
        self.add_results("QCP_simple.json")



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.config import ConfigValue, document_kwargs_from_configdict
from pyomo.common.dependencies import numpy as np
from pyomo.common.gc_manager import PauseGC
from pyomo.common.numeric_types import native_numeric_types
from pyomo.core import Var

from pyomo.opt import WriterFactory
from pyomo.repn.parameterized import ParameterizedLinearRepnVisitor
from pyomo.repn.plugins.standard_form import (
    LinearStandardFormInfo,
    LinearStandardFormCompiler,
    _LinearStandardFormCompiler_impl,
)
from pyomo.util.config_domains import ComponentDataSet


@WriterFactory.register(
    'compile_parameterized_standard_form',
    'Compile an LP to standard form (`min cTx s.t. Ax <= b`) treating some '
    'variables as data (e.g., variables decided by the outer problem in a '
    'bilevel optimization problem).',
)
class ParameterizedLinearStandardFormCompiler(LinearStandardFormCompiler):
    r"""Compiler to convert a "Parameterized" LP to the matrix representation
    of the standard form:

    .. math::

        \min\ & c^Tx \\
        s.t.\ & Ax \le b

    by treating the variables specified in the ``wrt`` list as data
    (constants).  The resulting compiled representation is returned as
    NumPy arrays and SciPy sparse matrices in a
    :py:class:`LinearStandardFormInfo` .

    """

    CONFIG = LinearStandardFormCompiler.CONFIG()
    CONFIG.declare(
        'wrt',
        ConfigValue(
            default=None,
            domain=ComponentDataSet(Var),
            description="Vars to treat as data for the purposes of compiling "
            "the standard form",
            doc="""
            Optional list of Vars to be treated as data while compiling the 
            standard form.

            For example, if this is the standard form of an inner problem in a
            multilevel optimization problem, then the outer problem's Vars would
            be specified in this list since they are not variables from the
            perspective of the inner problem.
            """,
        ),
    )

    @document_kwargs_from_configdict(CONFIG)
    def write(self, model, ostream=None, **options):
        r"""Convert a model to standard form treating the Vars specified in
        ``wrt`` as data.

        Returns
        -------
        LinearStandardFormInfo

        Parameters
        ----------
        model: ConcreteModel
            The concrete Pyomo model to write out.

        ostream: None
            This is provided for API compatibility with other writers
            and is ignored here.

        """
        config = self.config(options)

        # Pause the GC, as the walker that generates the compiled LP
        # representation generates (and disposes of) a large number of
        # small objects.
        with PauseGC():
            return _ParameterizedLinearStandardFormCompiler_impl(config).write(model)


class _SparseMatrixBase:
    def __init__(self, matrix_data, shape):
        data, indices, indptr = matrix_data
        nrows, ncols = shape

        self.data = np.array(data)
        self.indices = np.array(indices, dtype=int)
        self.indptr = np.array(indptr, dtype=int)
        self.shape = (nrows, ncols)

    def __eq__(self, other):
        return self.todense() == other


class _CSRMatrix(_SparseMatrixBase):
    def __init__(self, matrix_data, shape):
        super().__init__(matrix_data, shape)
        if len(self.indptr) != self.shape[0] + 1:
            raise ValueError(
                "Shape specifies the number of rows as %s but the index "
                "pointer has length %s. The index pointer must have length "
                "nrows + 1: Check the 'shape' and 'matrix_data' arguments."
                % (self.shape[0], len(self.indptr))
            )

    def tocsc(self):
        """Implements the same algorithm as scipy's csr_tocsc function from
        sparsetools.
        """
        csr_data = self.data
        col_index = self.indices
        row_index_ptr = self.indptr
        nrows = self.shape[0]

        num_nonzeros = len(csr_data)
        csc_data = np.empty(csr_data.shape[0], dtype=object)
        row_index = np.empty(num_nonzeros, dtype=int)
        # tally the nonzeros in each column
        col_index_ptr = np.zeros(self.shape[1], dtype=int)
        for i in col_index:
            col_index_ptr[int(i)] += 1

        # cumulative sum the tally to get the column index pointer
        cum_sum = 0
        for i, tally in enumerate(col_index_ptr):
            col_index_ptr[i] = cum_sum
            cum_sum += tally
        # We have now initialized the col_index_ptr to the *starting* position
        # of each column in the data vector.  Note that col_index_ptr is only
        # num_cols long: we have ignored the last entry in the standard CSC
        # col_index_ptr (the total number of nonzeros).  This will get resolved
        # below when we shift this vector by one position.

        # Now we are actually going to mess up what we just did while we
        # construct the row index: We can imagine that col_index_ptr holds the
        # position of the *next* nonzero in each column, so each time we move a
        # data element into a column we will increment that col_index_ptr by
        # one.  This is beautiful because by "messing up" the col_index_pointer,
        # we are just transforming the vector of *starting* indices for each
        # column to a vector of *ending* indices (actually 1 past the last
        # index) of each column. Thank you, scipy.
        for row in range(nrows):
            for j in range(row_index_ptr[row], row_index_ptr[row + 1]):
                col = col_index[j]
                dest = col_index_ptr[col]
                row_index[dest] = row
                # Note that the data changes order because now we are looking
                # for nonzeros through the columns rather than through the rows.
                csc_data[dest] = csr_data[j]

                col_index_ptr[col] += 1

        # Fix the column index pointer by inserting 0 at the beginning. The
        # col_index_ptr currently holds pointers to 1 past the last element of
        # each column, which is really the starting index for the next
        # column. Inserting the 0 (the starting index for the firsst column)
        # shifts everything by one column, "converting" the vector to the
        # starting indices of each column, and extending the vector length to
        # num_cols + 1 (as is expected by the CSC matrix).
        col_index_ptr = np.insert(col_index_ptr, 0, 0)

        return _CSCMatrix((csc_data, row_index, col_index_ptr), self.shape)

    def todense(self):
        """Implements the algorithm from scipy's csr_todense function
        in sparsetools.
        """
        nrows = self.shape[0]
        col_index = self.indices
        row_index_ptr = self.indptr
        data = self.data

        dense = np.zeros(self.shape, dtype=object)

        for row in range(nrows):
            for j in range(row_index_ptr[row], row_index_ptr[row + 1]):
                dense[row, col_index[j]] = data[j]

        return dense


class _CSCMatrix(_SparseMatrixBase):
    def __init__(self, matrix_data, shape):
        super().__init__(matrix_data, shape)
        if len(self.indptr) != self.shape[1] + 1:
            raise ValueError(
                "Shape specifies the number of columns as %s but the index "
                "pointer has length %s. The index pointer must have length "
                "ncols + 1: Check the 'shape' and 'matrix_data' arguments."
                % (self.shape[1], len(self.indptr))
            )

    def todense(self):
        """Implements the algorithm from scipy's csr_todense function
        in sparsetools.
        """
        ncols = self.shape[1]
        row_index = self.indices
        col_index_ptr = self.indptr
        data = self.data

        dense = np.zeros(self.shape, dtype=object)

        for col in range(ncols):
            for j in range(col_index_ptr[col], col_index_ptr[col + 1]):
                dense[row_index[j], col] = data[j]

        return dense

    def tocsr(self):
        """Implements the same algorithm as scipy's csr_tocsc function from
        sparsetools.
        """
        csc_data = self.data
        row_index = self.indices
        col_index_ptr = self.indptr
        ncols = self.shape[1]

        num_nonzeros = len(csc_data)
        csr_data = np.empty(csc_data.shape[0], dtype=object)
        col_index = np.empty(num_nonzeros, dtype=int)
        # tally the nonzeros in each column
        row_index_ptr = np.zeros(self.shape[0], dtype=int)
        for i in row_index:
            row_index_ptr[int(i)] += 1

        # cumulative sum the tally to get the column index pointer
        cum_sum = 0
        for i, tally in enumerate(row_index_ptr):
            row_index_ptr[i] = cum_sum
            cum_sum += tally
        # We have now initialized the row_index_ptr to the *starting* position
        # of each column in the data vector.  Note that row_index_ptr is only
        # num_rows long: we have ignored the last entry in the standard CSR
        # row_index_ptr (the total number of nonzeros).  This will get resolved
        # below when we shift this vector by one position.

        # Now we are actually going to mess up what we just did while we
        # construct the row index: We can imagine that row_index_ptr holds the
        # position of the *next* nonzero in each column, so each time we move a
        # data element into a column we will increment that row_index_ptr by
        # one.  This is beautiful because by "messing up" the row_index_pointer,
        # we are just transforming the vector of *starting* indices for each
        # column to a vector of *ending* indices (actually 1 past the last
        # index) of each column. Thank you, scipy.
        for col in range(ncols):
            for j in range(col_index_ptr[col], col_index_ptr[col + 1]):
                row = row_index[j]
                dest = row_index_ptr[row]
                col_index[dest] = col
                # Note that the data changes order because now we are looking
                # for nonzeros through the columns rather than through the rows.
                csr_data[dest] = csc_data[j]

                row_index_ptr[row] += 1

        # Fix the row index pointer by inserting 0 at the beginning. The
        # row_index_ptr currently holds pointers to 1 past the last element of
        # each row, which is really the starting index for the next
        # row. Inserting the 0 (the starting index for the first column)
        # shifts everything by one column, "converting" the vector to the
        # starting indices of each row, and extending the vector length to
        # num_rows + 1 (as is expected by the CSR matrix).
        row_index_ptr = np.insert(row_index_ptr, 0, 0)

        return _CSRMatrix((csr_data, col_index, row_index_ptr), self.shape)

    def sum_duplicates(self):
        """Implements the algorithm from scipy's csr_sum_duplicates function
        in sparsetools.

        Note that this only removes duplicates that are adjacent, so it will remove
        all duplicates if the incoming CSC matrix has sorted indices. (In particular
        this will be true if it was just converted from CSR).
        """
        ncols = self.shape[1]
        row_index = self.indices
        col_index_ptr = self.indptr
        data = self.data

        num_non_zeros = 0
        col_end = 0
        for i in range(ncols):
            jj = col_end
            col_end = col_index_ptr[i + 1]
            while jj < col_end:
                j = row_index[jj]
                x = data[jj]
                jj += 1
                while jj < col_end and row_index[jj] == j:
                    x += data[jj]
                    jj += 1
                row_index[num_non_zeros] = j
                data[num_non_zeros] = x
                num_non_zeros += 1
            col_index_ptr[i + 1] = num_non_zeros

        # [ESJ 11/11/24]: I'm not 100% sure how scipy handles this, but we need
        # to remove the "extra" entries from the data and row_index arrays.
        self.data = data[:num_non_zeros]
        self.row_index = row_index[:num_non_zeros]

    def eliminate_zeros(self):
        """Implements the algorithm from scipy's csr_eliminate_zeros function
        in sparsetools.
        """
        ncols = self.shape[1]
        row_index = self.indices
        col_index_ptr = self.indptr
        data = self.data

        num_non_zeros = 0
        col_end = 0
        for i in range(ncols):
            jj = col_end
            col_end = col_index_ptr[i + 1]
            while jj < col_end:
                j = row_index[jj]
                x = data[jj]
                if x.__class__ not in native_numeric_types or x != 0:
                    row_index[num_non_zeros] = j
                    data[num_non_zeros] = x
                    num_non_zeros += 1
                jj += 1
            col_index_ptr[i + 1] = num_non_zeros


class _ParameterizedLinearStandardFormCompiler_impl(_LinearStandardFormCompiler_impl):
    _csc_matrix = _CSCMatrix
    _csr_matrix = _CSRMatrix

    def _get_visitor(self, subexpression_cache, var_recorder):
        wrt = self.config.wrt
        if wrt is None:
            wrt = []
        return ParameterizedLinearRepnVisitor(
            subexpression_cache, wrt=wrt, var_recorder=var_recorder
        )

    def _to_vector(self, data, N, vector_type):
        # override this to not attempt conversion to float since that will fail
        # on the Pyomo expressions
        return np.array(list(data))



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core.base.reference import Reference
from pyomo.common.collections import ComponentMap
from pyomo.common.modeling import unique_component_name


def rename_components(model, component_list, prefix):
    """Rename components in component_list using the prefix AND
    unique_component_name

    Parameters
    ----------
    model : Pyomo model (or Block)
       The variables, constraints and objective will be renamed on this model
    component_list : list
       List of components to rename
    prefix : str
       The prefix to use when building the new names

    Examples
    --------
    >>> model = pyo.ConcreteModel()
    >>> model.x = pyo.Var()
    >>> model.y = pyo.Var()
    >>> c_list = list(model.component_objects(ctype=pyo.Var, descend_into=True))
    >>> new = rename_components(model, component_list=c_list, prefix='special_')
    >>> str(new)
    "ComponentMap({'special_x (key=...)': 'x', 'special_y (key=...)': 'y'})"

    Returns
    -------
    ComponentMap : maps the renamed Component objects
       to a string that provides their old fully qualified names

    ToDo
    ----
    - need to add a check to see if someone accidentally passes a
      generator since this can lead to an infinite loop

    """
    # Need to collect any Reference first so that we can record the old mapping of data objects before renaming
    refs = {}
    for c in component_list:
        if c.is_reference():
            refs[c] = {}
            for k, v in c._data.items():
                refs[c][k] = (v.parent_block(), v.local_name)

    # Now rename all the non-Reference components
    name_map = ComponentMap()
    for c in component_list:
        # get the parent block - we will use this to ensure new names are
        # unique and to add the new "scaled" components
        if not c.is_reference():
            # Skip References for now
            parent = c.parent_block()
            old_name = c.name
            new_name = unique_component_name(parent, prefix + c.local_name)
            parent.del_component(c)
            parent.add_component(new_name, c)
            name_map[c] = old_name

    # Finally, replace all the References with remapped equivalents
    for c in refs:
        # Get mapping of new component names
        new_map = ComponentMap()
        for k, v in refs[c].items():
            # Assume the data was renamed
            new_data = v[0].find_component(prefix + v[1])

            if new_data is None:
                # If we couldn't find the new name, try the old one
                # This might happen if the reference pointed to something outside
                # the scope of the renaming
                new_data = v[0].find_component(v[1])

            if new_data is None:
                # If we still haven't found it, give up
                raise RuntimeError(
                    f"Unable to remap Reference {c.name} whilst renaming components."
                )

            new_map[k] = new_data

        parent = c.parent_block()
        old_name = c.name
        new_name = unique_component_name(parent, prefix + c.local_name)
        parent.del_component(c)
        cnew = Reference(new_map)
        parent.add_component(new_name, cnew)
        name_map[cnew] = old_name

    return name_map


def iter_component(obj):
    """Yield "child" objects from a component that is defined with either
    the `base` or `kernel` APIs.  If the component is not indexed, it
    returns itself.

    Parameters
    ----------
    obj : ComponentType
        eg. `TupleContainer`, `ListContainer`, `DictContainer`,
        `IndexedComponent`, or `Component`

    Returns
    -------
    Iterator[ComponentType] : Iterator of the component data objects.

    """
    try:
        # catches `IndexedComponent`, and kernel's `_dict`
        return iter(obj.values())
    except AttributeError:
        pass

    try:
        # catches list and kernel's `_list` and `_tuple`
        return iter(obj)
    except TypeError:
        return iter((obj,))



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

#
# declare deprecation paths for removed modules
#
from pyomo.common.deprecation import moved_module

moved_module(
    'pyomo.repn.tests.ampl.nl_diff',
    'pyomo.repn.tests.nl_diff',
    version='6.6.0',
    remove_in='6.6.1',
)
del moved_module



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.network import Port, Arc
from pyomo.network.foqus_graph import FOQUSGraph
from pyomo.core import (
    Constraint,
    value,
    Objective,
    Var,
    ConcreteModel,
    Binary,
    minimize,
    Expression,
)
from pyomo.common.collections import ComponentSet, ComponentMap, Bunch
from pyomo.core.expr import identify_variables
from pyomo.repn import generate_standard_repn
import logging, time

from pyomo.common.dependencies import (
    networkx as nx,
    networkx_available,
    numpy,
    numpy_available,
)

imports_available = networkx_available & numpy_available

logger = logging.getLogger('pyomo.network')


class SequentialDecomposition(FOQUSGraph):
    """
    A sequential decomposition tool for Pyomo Network models

    The following parameters can be set upon construction of this class
    or via the `options` attribute.

    Parameters
    ----------
        graph: `MultiDiGraph`
            A networkx graph representing the model to be solved.

            `default=None (will compute it)`

        tear_set: `list`
            A list of indexes representing edges to be torn. Can be set with
            a list of edge tuples via set_tear_set.

            `default=None (will compute it)`

        select_tear_method: `str`
            Which method to use to select a tear set, either "mip" or
            "heuristic".

            `default="mip"`

        run_first_pass: `bool`
            Boolean indicating whether or not to run through network before
            running the tear stream convergence procedure.

            `default=True`

        solve_tears: `bool`
            Boolean indicating whether or not to run iterations to converge
            tear streams.

            `default=True`

        guesses: `ComponentMap`
            ComponentMap of guesses to use for first pass
            (see set_guesses_for method).

            `default=ComponentMap()`

        default_guess: `float`
            Value to use if a free variable has no guess.

            `default=None`

        almost_equal_tol: `float`
            Difference below which numbers are considered equal when checking
            port value agreement.

            `default=1.0E-8`

        log_info: `bool`
            Set logger level to INFO during run.

            `default=False`

        tear_method: `str`
            Method to use for converging tear streams, either "Direct" or
            "Wegstein".

            `default="Direct"`

        iterLim: `int`
            Limit on the number of tear iterations.

            `default=40`

        tol: `float`
            Tolerance at which to stop tear iterations.

            `default=1.0E-5`

        tol_type: `str`
            Type of tolerance value, either "abs" (absolute) or
            "rel" (relative to current value).

            `default="abs"`

        report_diffs: `bool`
            Report the matrix of differences across tear streams for
            every iteration.

            `default=False`

        accel_min: `float`
            Min value for Wegstein acceleration factor.

            `default=-5`

        accel_max: `float`
            Max value for Wegstein acceleration factor.

            `default=0`

        tear_solver: `str`
            Name of solver to use for select_tear_mip.

            `default="cplex"`

        tear_solver_io: `str`
            Solver IO keyword for the above solver.

            `default=None`

        tear_solver_options: `dict`
            Keyword options to pass to solve method.

            `default={}`
    """

    def __init__(self, **kwds):
        """Pass kwds to update the options attribute after setting defaults"""
        self.cache = {}
        options = self.options = Bunch()
        # defaults
        options["graph"] = None
        options["tear_set"] = None
        options["select_tear_method"] = "mip"
        options["run_first_pass"] = True
        options["solve_tears"] = True
        options["guesses"] = ComponentMap()
        options["default_guess"] = None
        options["almost_equal_tol"] = 1.0e-8
        options["log_info"] = False
        options["tear_method"] = "Direct"
        options["iterLim"] = 40
        options["tol"] = 1.0e-5
        options["tol_type"] = "abs"
        options["report_diffs"] = False
        options["accel_min"] = -5
        options["accel_max"] = 0
        options["tear_solver"] = "cplex"
        options["tear_solver_io"] = None
        options["tear_solver_options"] = {}

        options.update(kwds)

    def set_guesses_for(self, port, guesses):
        """
        Set the guesses for the given port

        These guesses will be checked for all free variables that are
        encountered during the first pass run. If a free variable has
        no guess, its current value will be used. If its current value
        is None, the default_guess option will be used. If that is None,
        an error will be raised.

        All port variables that are downstream of a non-tear edge will
        already be fixed. If there is a guess for a fixed variable, it
        will be silently ignored.

        The guesses should be a dict that maps the following:

            Port Member Name -> Value

        Or, for indexed members, multiple dicts that map:

            Port Member Name -> Index -> Value

        For extensive members, "Value" must be a list of tuples of the
        form (arc, value) to guess a value for the expanded variable
        of the specified arc. However, if the arc connecting this port
        is a 1-to-1 arc with its peer, then there will be no expanded
        variable for the single arc, so a regular "Value" should be
        provided.

        This dict cannot be used to pass guesses for variables within
        expression type members. Guesses for those variables must be
        assigned to the variable's current value before calling run.

        While this method makes things more convenient, all it does is:

            `self.options["guesses"][port] = guesses`
        """
        self.options["guesses"][port] = guesses

    def set_tear_set(self, tset):
        """
        Set a custom tear set to be used when running the decomposition

        The procedure will use this custom tear set instead of finding
        its own, thus it can save some time. Additionally, this will be
        useful for knowing which edges will need guesses.

        Arguments
        ---------
            tset
                A list of Arcs representing edges to tear

        While this method makes things more convenient, all it does is:

            `self.options["tear_set"] = tset`
        """
        self.options["tear_set"] = tset

    def tear_set_arcs(self, G, method="mip", **kwds):
        """
        Call the specified tear selection method and return a list
        of arcs representing the selected tear edges.

        The kwds will be passed to the method.
        """
        if method == "mip":
            tset = self.select_tear_mip(G, **kwds)
        elif method == "heuristic":
            # tset is the first list in the first return value
            tset = self.select_tear_heuristic(G, **kwds)[0][0]
        else:
            raise ValueError("Invalid method '%s'" % (method,))

        return self.indexes_to_arcs(G, tset)

    def indexes_to_arcs(self, G, lst):
        """
        Converts a list of edge indexes to the corresponding Arcs

        Arguments
        ---------
            G
                A networkx graph corresponding to lst
            lst
                A list of edge indexes to convert to tuples

        Returns:
            A list of arcs
        """
        edge_list = self.idx_to_edge(G)
        res = []
        for ei in lst:
            edge = edge_list[ei]
            res.append(G.edges[edge]["arc"])
        return res

    def run(self, model, function):
        """
        Compute a Pyomo Network model using sequential decomposition

        Arguments
        ---------
            model
                A Pyomo model
            function
                A function to be called on each block/node in the network
        """
        if self.options["log_info"]:
            old_log_level = logger.level
            logger.setLevel(logging.INFO)

        self.cache.clear()

        try:
            return self._run_impl(model, function)
        finally:
            # Cleanup
            self.cache.clear()

            if self.options["log_info"]:
                logger.setLevel(old_log_level)

    def _run_impl(self, model, function):
        start = time.time()
        logger.info("Starting Sequential Decomposition")

        G = self.options["graph"]
        if G is None:
            G = self.create_graph(model)

        tset = self.tear_set(G)

        if self.options["run_first_pass"]:
            logger.info("Starting first pass run of network")
            order = self.calculation_order(G)
            self.run_order(G, order, function, tset, use_guesses=True)

        if not self.options["solve_tears"] or not len(tset):
            # Not solving tears, we're done
            end = time.time()
            logger.info(
                "Finished Sequential Decomposition in %.2f seconds" % (end - start)
            )
            return

        logger.info("Starting tear convergence procedure")

        sccNodes, sccEdges, sccOrder, outEdges = self.scc_collect(G)

        for lev in sccOrder:
            for sccIndex in lev:
                order = self.calculation_order(G, nodes=sccNodes[sccIndex])

                # only pass tears that are part of this SCC
                tears = []
                for ei in tset:
                    if ei in sccEdges[sccIndex]:
                        tears.append(ei)

                kwds = dict(
                    G=G,
                    order=order,
                    function=function,
                    tears=tears,
                    iterLim=self.options["iterLim"],
                    tol=self.options["tol"],
                    tol_type=self.options["tol_type"],
                    report_diffs=self.options["report_diffs"],
                    outEdges=outEdges[sccIndex],
                )

                tear_method = self.options["tear_method"]

                if tear_method == "Direct":
                    self.solve_tear_direct(**kwds)

                elif tear_method == "Wegstein":
                    kwds["accel_min"] = self.options["accel_min"]
                    kwds["accel_max"] = self.options["accel_max"]
                    self.solve_tear_wegstein(**kwds)

                else:
                    raise ValueError("Invalid tear_method '%s'" % (tear_method,))

        end = time.time()
        logger.info("Finished Sequential Decomposition in %.2f seconds" % (end - start))

    def run_order(self, G, order, function, ignore=None, use_guesses=False):
        """
        Run computations in the order provided by calling the function

        Arguments
        ---------
            G
                A networkx graph corresponding to order
            order
                The order in which to run each node in the graph
            function
                The function to be called on each block/node
            ignore
                Edge indexes to ignore when passing values
            use_guesses
                If True, will check the guesses dict when fixing
                free variables before calling function
        """
        fixed_inputs = self.fixed_inputs()
        fixed_outputs = ComponentSet()
        edge_map = self.edge_to_idx(G)
        arc_map = self.arc_to_edge(G)
        guesses = self.options["guesses"]
        default = self.options["default_guess"]
        for lev in order:
            for unit in lev:
                if unit not in fixed_inputs:
                    fixed_inputs[unit] = ComponentSet()
                fixed_ins = fixed_inputs[unit]

                # make sure all inputs are fixed
                for port in unit.component_data_objects(Port):
                    if not len(port.sources()):
                        continue
                    if use_guesses and port in guesses:
                        self.load_guesses(guesses, port, fixed_ins)
                    self.load_values(port, default, fixed_ins, use_guesses)

                function(unit)

                # free the inputs that were not already fixed
                for var in fixed_ins:
                    var.free()
                fixed_ins.clear()

                # pass the values downstream for all outlet ports
                for port in unit.component_data_objects(Port):
                    dests = port.dests()
                    if not len(dests):
                        continue
                    for var in port.iter_vars(expr_vars=True, fixed=False):
                        fixed_outputs.add(var)
                        var.fix()
                    for arc in dests:
                        # arc might not be in the arc_map if
                        # it is on a deactivated block
                        if arc in arc_map and edge_map[arc_map[arc]] not in ignore:
                            self.pass_values(arc, fixed_inputs)
                    for var in fixed_outputs:
                        var.free()
                    fixed_outputs.clear()

    def pass_values(self, arc, fixed_inputs):
        """
        Pass the values from one unit to the next, recording only those that
        were not already fixed in the provided dict that maps blocks to sets.
        """
        eblock = arc.expanded_block
        src, dest = arc.src, arc.dest
        dest_unit = dest.parent_block()
        eq_tol = self.options["almost_equal_tol"]

        if dest_unit not in fixed_inputs:
            fixed_inputs[dest_unit] = ComponentSet()

        sf = eblock.component("splitfrac")
        if sf is not None and not sf.is_fixed():
            # fix the splitfrac if it has a current value or else error
            if sf.value is not None:
                fixed_inputs[dest_unit].add(sf)
                sf.fix()
            else:
                raise RuntimeError(
                    "Found free splitfrac for arc '%s' with no current value. "
                    "Please use the set_split_fraction method on its source "
                    "port to set this value before expansion, or set its value "
                    "manually if expansion has already occurred." % arc.name
                )
        elif sf is None:
            # if there is no splitfrac, but we have extensive members, then we
            # need to manually set the evar values because there will be no
            # *_split constraints on the eblock, so it is up to us to set it
            # TODO: what if there is no splitfrac, but it's missing because
            # there's only 1 variable per port so it was simplified out?
            # How do we specify the downstream evars? If we assume that the
            # user's function will satisfy the *_outsum constraint before
            # returning, then the evars would at least be specified such that
            # they satisfy the total sum constraint. But I would think we don't
            # want to rely on the user calling solve on their unit before
            # returning, especially since the outsum constraint was auto
            # generated and not one they made themselves.
            # Potential Solution: allow the user to specify a splitfrac
            # (via set_split_fraction or something else) that will be used here
            # and is only relevant to this SM, and if they didn't specify
            # anything, throw an error.
            for name, mem in src.vars.items():
                if not src.is_extensive(name):
                    continue
                evar = eblock.component(name)
                if evar is None:
                    continue
                if len(src.dests()) > 1:
                    raise Exception(
                        "This still needs to be figured out (arc '%s')" % arc.name
                    )
                # TODO: for now we know it's obvious what to do if there is
                # only 1 destination
                if mem.is_indexed():
                    evars = [(evar[i], i) for i in evar]
                else:
                    evars = [(evar, None)]
                for evar, idx in evars:
                    fixed_inputs[dest_unit].add(evar)
                    val = value(mem[idx] if mem.is_indexed() else mem)
                    # val are numpy.float64; coerce val back to float
                    evar.fix(float(val))

        for con in eblock.component_data_objects(Constraint, active=True):
            # we expect to find equality constraints with one linear variable
            if not con.equality:
                # We assume every constraint here is an equality.
                # This will only be False if the transformation changes
                # or if the user puts something unexpected on the eblock.
                raise RuntimeError(
                    "Found inequality constraint '%s'. Please do not modify "
                    "the expanded block." % con.name
                )
            repn = generate_standard_repn(con.body)
            if repn.is_fixed():
                # the port member's peer was already fixed
                if abs(value(con.lower) - repn.constant) > eq_tol:
                    raise RuntimeError(
                        "Found connected ports '%s' and '%s' both with fixed "
                        "but different values (by > %s) for constraint '%s'"
                        % (src, dest, eq_tol, con.name)
                    )
                continue
            if not (repn.is_linear() and len(repn.linear_vars) == 1):
                raise RuntimeError(
                    "Constraint '%s' had more than one free variable when "
                    "trying to pass a value to its destination. Please fix "
                    "more variables before passing across this arc." % con.name
                )
            # fix the value of the single variable to satisfy the constraint
            # con.lower is usually a NumericConstant but call value on it
            # just in case it is something else
            val = (value(con.lower) - repn.constant) / repn.linear_coefs[0]
            var = repn.linear_vars[0]
            fixed_inputs[dest_unit].add(var)
            # val are numpy.float64; coerce val back to float
            var.fix(float(val))

    def pass_single_value(self, port, name, member, val, fixed):
        """
        Fix the value of the port member and add it to the fixed set.
        If the member is an expression, appropriately fix the value of
        its free variable. Error if the member is already fixed but
        different from val, or if the member has more than one free
        variable."
        """
        eq_tol = self.options["almost_equal_tol"]
        if member.is_fixed():
            if abs(value(member) - val) > eq_tol:
                raise RuntimeError(
                    "Member '%s' of port '%s' is already fixed but has a "
                    "different value (by > %s) than what is being passed to it"
                    % (name, port.name, eq_tol)
                )
        elif member.is_expression_type():
            repn = generate_standard_repn(member - val)
            if repn.is_linear() and len(repn.linear_vars) == 1:
                # fix the value of the single variable
                fval = (0 - repn.constant) / repn.linear_coefs[0]
                var = repn.linear_vars[0]
                fixed.add(var)
                # val are numpy.float64; coerce val back to float
                var.fix(float(fval))
            else:
                raise RuntimeError(
                    "Member '%s' of port '%s' had more than "
                    "one free variable when trying to pass a value "
                    "to it. Please fix more variables before passing "
                    "to this port." % (name, port.name)
                )
        else:
            fixed.add(member)
            # val are numpy.float64; coerce val back to float
            member.fix(float(val))

    def load_guesses(self, guesses, port, fixed):
        srcs = port.sources()
        for name, mem in port.vars.items():
            try:
                entry = guesses[port][name]
            except KeyError:
                continue

            if isinstance(entry, dict):
                itr = [(mem[k], entry[k], k) for k in entry]
            elif mem.is_indexed():
                raise TypeError(
                    "Guess for indexed member '%s' in port '%s' must map to a "
                    "dict of indexes" % (name, port.name)
                )
            else:
                itr = [(mem, entry, None)]

            for var, entry, idx in itr:
                if var.is_fixed():
                    # silently ignore vars already fixed
                    continue
                has_evars = False
                if port.is_extensive(name):
                    for arc, val in entry:
                        if arc not in srcs:
                            raise ValueError(
                                "Found a guess for extensive member '%s' on "
                                "port '%s' using arc '%s' that is not a source "
                                "of this port" % (name, port.name, arc.name)
                            )
                        evar = arc.expanded_block.component(name)
                        if evar is None:
                            # no evars, 1-to-1 arc
                            break
                        has_evars = True
                        # even if idx is None, we know evar is a Var and
                        # indexing by None into ScalarVars returns itself
                        evar = evar[idx]
                        if evar.is_fixed():
                            # silently ignore vars already fixed
                            continue
                        fixed.add(evar)
                        evar.fix(float(val))
                if not has_evars:
                    # the only NumericValues in Pyomo that return True
                    # for is_fixed are expressions and variables
                    if var.is_expression_type():
                        raise ValueError(
                            "Cannot provide guess for expression type member "
                            "'%s%s' of port '%s', must set current value of "
                            "variables within expression"
                            % (
                                name,
                                ("[%s]" % str(idx)) if mem.is_indexed() else "",
                                port.name,
                            )
                        )
                    else:
                        fixed.add(var)
                        var.fix(float(entry))

    def load_values(self, port, default, fixed, use_guesses):
        sources = port.sources()
        for name, index, obj in port.iter_vars(fixed=False, names=True):
            evars = None
            if port.is_extensive(name):
                # collect evars if there are any
                evars = [arc.expanded_block.component(name) for arc in sources]
                if evars[0] is None:
                    # no evars, so this arc is 1-to-1
                    evars = None
                else:
                    try:
                        # index into them if necessary, now that
                        # we know they are not None
                        for j in range(len(evars)):
                            evars[j] = evars[j][index]
                    except AttributeError:
                        pass
            if evars is not None:
                for evar in evars:
                    if evar.is_fixed():
                        continue
                    self.check_value_fix(
                        port, evar, default, fixed, use_guesses, extensive=True
                    )
                # now all evars should be fixed so combine them
                # and fix the value of the extensive port member
                self.combine_and_fix(port, name, obj, evars, fixed)
            else:
                if obj.is_expression_type():
                    for var in identify_variables(obj, include_fixed=False):
                        self.check_value_fix(port, var, default, fixed, use_guesses)
                else:
                    self.check_value_fix(port, obj, default, fixed, use_guesses)

    def check_value_fix(self, port, var, default, fixed, use_guesses, extensive=False):
        """
        Try to fix the var at its current value or the default, else error
        """
        val = None
        if var.value is not None:
            val = var.value
        elif default is not None:
            val = default

        if val is None:
            raise RuntimeError(
                "Encountered a free inlet %svariable '%s' %s port '%s' with no "
                "%scurrent value, or default_guess option, while attempting "
                "to compute the unit."
                % (
                    "extensive " if extensive else "",
                    var.name,
                    ("on", "to")[int(extensive)],
                    port.name,
                    "guess, " if use_guesses else "",
                )
            )

        fixed.add(var)
        var.fix(float(val))

    def combine_and_fix(self, port, name, obj, evars, fixed):
        """
        For an extensive port member, combine the values of all
        expanded variables and fix the port member at their sum.
        Assumes that all expanded variables are fixed.
        """
        assert all(evar.is_fixed() for evar in evars)
        total = sum(value(evar) for evar in evars)
        self.pass_single_value(port, name, obj, total, fixed)

    def source_dest_peer(self, arc, name, index=None):
        """
        Return the object that is the peer to the source port's member.
        This is either the destination port's member, or the variable
        on the arc's expanded block for Extensive properties. This will
        return the appropriate index of the peer.
        """
        # check the rule on source but dest should be the same
        if arc.src.is_extensive(name):
            evar = arc.expanded_block.component(name)
            if evar is not None:
                # 1-to-1 arcs don't make evar because they're an equality
                return evar[index]
        mem = arc.dest.vars[name]
        if mem.is_indexed():
            return mem[index]
        else:
            return mem

    def create_graph(self, model):
        """
        Returns a networkx MultiDiGraph of a Pyomo network model

        The nodes are units and the edges follow Pyomo Arc objects. Nodes
        that get added to the graph are determined by the parent blocks
        of the source and destination Ports of every Arc in the model.
        Edges are added for each Arc using the direction specified by
        source and destination. All Arcs in the model will be used whether
        or not they are active (since this needs to be done after expansion),
        and they all need to be directed.
        """
        G = nx.MultiDiGraph()

        for blk in model.block_data_objects(descend_into=True, active=True):
            for arc in blk.component_data_objects(Arc, descend_into=False):
                if not arc.directed:
                    raise ValueError(
                        "All Arcs must be directed when creating "
                        "a graph for a model. Found undirected "
                        "Arc: '%s'" % arc.name
                    )
                if arc.expanded_block is None:
                    raise ValueError(
                        "All Arcs must be expanded when creating "
                        "a graph for a model. Found unexpanded "
                        "Arc: '%s'" % arc.name
                    )
                src, dest = arc.src.parent_block(), arc.dest.parent_block()
                G.add_edge(src, dest, arc=arc)

        return G

    def select_tear_mip_model(self, G):
        """
        Generate a model for selecting tears from the given graph

        Returns
        -------
            model
            bin_list
                A list of the binary variables representing each edge,
                indexed by the edge index of the graph
        """
        model = ConcreteModel()

        bin_list = []
        for i in range(G.number_of_edges()):
            # add a binary "torn" variable for every edge
            vname = "edge%s" % i
            var = Var(domain=Binary)
            bin_list.append(var)
            model.add_component(vname, var)

        # var containing the maximum number of times any cycle is torn
        mct = model.max_cycle_tears = Var()

        _, cycleEdges = self.all_cycles(G)

        for i in range(len(cycleEdges)):
            ecyc = cycleEdges[i]

            # expression containing sum of tears for each cycle
            ename = "cycle_sum%s" % i
            expr = Expression(expr=sum(bin_list[i] for i in ecyc))
            model.add_component(ename, expr)

            # every cycle must have at least 1 tear
            cname_min = "cycle_min%s" % i
            con_min = Constraint(expr=expr >= 1)
            model.add_component(cname_min, con_min)

            # mct >= cycle_sum for all cycles, thus it becomes the max
            cname_mct = mct.name + "_geq%s" % i
            con_mct = Constraint(expr=mct >= expr)
            model.add_component(cname_mct, con_mct)

        # weigh the primary objective much greater than the secondary
        obj_expr = 1000 * mct + sum(var for var in bin_list)
        model.obj = Objective(expr=obj_expr, sense=minimize)

        return model, bin_list

    def select_tear_mip(self, G, solver, solver_io=None, solver_options={}):
        """
        This finds optimal sets of tear edges based on two criteria.
        The primary objective is to minimize the maximum number of
        times any cycle is broken. The secondary criteria is to
        minimize the number of tears.

        This function creates a MIP problem in Pyomo with a doubly
        weighted objective and solves it with the solver arguments.
        """
        model, bin_list = self.select_tear_mip_model(G)

        from pyomo.environ import SolverFactory

        opt = SolverFactory(solver, solver_io=solver_io)
        if not opt.available(exception_flag=False):
            raise ValueError(
                "Solver '%s' (solver_io=%r) is not available, please pass a "
                "different solver" % (solver, solver_io)
            )
        opt.solve(model, **solver_options)

        # collect final list by adding every edge with a "True" binary var
        tset = []
        for i in range(len(bin_list)):
            if bin_list[i].value == 1:
                tset.append(i)

        return tset

    def compute_err(self, svals, dvals, tol_type):
        """Compute the diff between svals and dvals for the given tol_type"""
        if tol_type not in ("abs", "rel"):
            raise ValueError("Invalid tol_type '%s'" % (tol_type,))

        diff = svals - dvals
        if tol_type == "abs":
            err = diff
        else:
            # relative: divide by current value of svals
            old_settings = numpy.seterr(divide='ignore', invalid='ignore')
            err = diff / svals
            numpy.seterr(**old_settings)
            # isnan means 0/0 so diff is 0
            err[numpy.isnan(err)] = 0
            # isinf means diff/0, so just use the diff
            if any(numpy.isinf(err)):
                for i in range(len(err)):
                    if numpy.isinf(err[i]):
                        err[i] = diff[i]

        return err

    def tear_diff_direct(self, G, tears):
        """
        Returns numpy arrays of values for src and dest members
        for all edges in the tears list of edge indexes.
        """
        svals = []
        dvals = []
        edge_list = self.idx_to_edge(G)
        for tear in tears:
            arc = G.edges[edge_list[tear]]["arc"]
            src, dest = arc.src, arc.dest
            sf = arc.expanded_block.component("splitfrac")
            for name, index, mem in src.iter_vars(names=True):
                if src.is_extensive(name) and sf is not None:
                    # TODO: same as above, what if there's no splitfrac
                    svals.append(value(mem * sf))
                else:
                    svals.append(value(mem))
                dvals.append(value(self.source_dest_peer(arc, name, index)))
        svals = numpy.array(svals)
        dvals = numpy.array(dvals)
        return svals, dvals

    def pass_edges(self, G, edges):
        """Call pass values for a list of edge indexes"""
        fixed_outputs = ComponentSet()
        edge_list = self.idx_to_edge(G)
        for ei in edges:
            arc = G.edges[edge_list[ei]]["arc"]
            for var in arc.src.iter_vars(expr_vars=True, fixed=False):
                fixed_outputs.add(var)
                var.fix()
            self.pass_values(arc, self.fixed_inputs())
            for var in fixed_outputs:
                var.free()
            fixed_outputs.clear()

    def pass_tear_direct(self, G, tears):
        """Pass values across all tears in the given tear set"""
        fixed_outputs = ComponentSet()
        edge_list = self.idx_to_edge(G)

        for tear in tears:
            # fix everything then call pass values
            arc = G.edges[edge_list[tear]]["arc"]
            for var in arc.src.iter_vars(expr_vars=True, fixed=False):
                fixed_outputs.add(var)
                var.fix()
            self.pass_values(arc, fixed_inputs=self.fixed_inputs())
            for var in fixed_outputs:
                var.free()
            fixed_outputs.clear()

    def pass_tear_wegstein(self, G, tears, x):
        """
        Set the destination value of all tear edges to
        the corresponding value in the numpy array x.
        """
        fixed_inputs = self.fixed_inputs()
        edge_list = self.idx_to_edge(G)
        i = 0
        for tear in tears:
            arc = G.edges[edge_list[tear]]["arc"]
            src, dest = arc.src, arc.dest
            dest_unit = dest.parent_block()

            if dest_unit not in fixed_inputs:
                fixed_inputs[dest_unit] = ComponentSet()

            for name, index, mem in src.iter_vars(names=True):
                peer = self.source_dest_peer(arc, name, index)
                self.pass_single_value(dest, name, peer, x[i], fixed_inputs[dest_unit])
                i += 1

    def generate_gofx(self, G, tears):
        edge_list = self.idx_to_edge(G)
        gofx = []
        for tear in tears:
            arc = G.edges[edge_list[tear]]["arc"]
            src = arc.src
            sf = arc.expanded_block.component("splitfrac")
            for name, index, mem in src.iter_vars(names=True):
                if src.is_extensive(name) and sf is not None:
                    # TODO: same as above, what if there's no splitfrac
                    gofx.append(value(mem * sf))
                else:
                    gofx.append(value(mem))
        gofx = numpy.array(gofx)
        return gofx

    def generate_first_x(self, G, tears):
        edge_list = self.idx_to_edge(G)
        x = []
        for tear in tears:
            arc = G.edges[edge_list[tear]]["arc"]
            for name, index, mem in arc.src.iter_vars(names=True):
                peer = self.source_dest_peer(arc, name, index)
                x.append(value(peer))
        x = numpy.array(x)
        return x

    def cacher(self, key, fcn, *args):
        if key in self.cache:
            return self.cache[key]
        res = fcn(*args)
        self.cache[key] = res
        return res

    def tear_set(self, G):
        key = "tear_set"

        def fcn(G):
            tset = self.options[key]
            if tset is not None:
                arc_map = self.arc_to_edge(G)
                edge_map = self.edge_to_idx(G)
                res = []
                for arc in tset:
                    res.append(edge_map[arc_map[arc]])
                if not self.check_tear_set(G, res):
                    raise ValueError(
                        "Tear set found in options is insufficient to solve network"
                    )
                self.cache[key] = res
                return res

            method = self.options["select_tear_method"]
            if method == "mip":
                return self.select_tear_mip(
                    G,
                    self.options["tear_solver"],
                    self.options["tear_solver_io"],
                    self.options["tear_solver_options"],
                )
            elif method == "heuristic":
                # tset is the first list in the first return value
                return self.select_tear_heuristic(G)[0][0]
            else:
                raise ValueError("Invalid select_tear_method '%s'" % (method,))

        return self.cacher(key, fcn, G)

    def arc_to_edge(self, G):
        """Returns a mapping from arcs to edges for a graph"""

        def fcn(G):
            res = ComponentMap()
            for edge in G.edges:
                arc = G.edges[edge]["arc"]
                res[arc] = edge
            return res

        return self.cacher("arc_to_edge", fcn, G)

    def fixed_inputs(self):
        return self.cacher("fixed_inputs", dict)

    def idx_to_node(self, G):
        """Returns a mapping from indexes to nodes for a graph"""
        return self.cacher("idx_to_node", list, G.nodes)

    def node_to_idx(self, G):
        """Returns a mapping from nodes to indexes for a graph"""

        def fcn(G):
            res = dict()
            i = -1
            for node in G.nodes:
                i += 1
                res[node] = i
            return res

        return self.cacher("node_to_idx", fcn, G)

    def idx_to_edge(self, G):
        """Returns a mapping from indexes to edges for a graph"""
        return self.cacher("idx_to_edge", list, G.edges)

    def edge_to_idx(self, G):
        """Returns a mapping from edges to indexes for a graph"""

        def fcn(G):
            res = dict()
            i = -1
            for edge in G.edges:
                i += 1
                res[edge] = i
            return res

        return self.cacher("edge_to_idx", fcn, G)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging

from pyomo.core.expr import inequality
from pyomo.core.base import (
    Transformation,
    TransformationFactory,
    Constraint,
    Block,
    SortComponents,
    ComponentUID,
)
from pyomo.mpec.complementarity import Complementarity
from pyomo.gdp.disjunct import Disjunct, Disjunction

logger = logging.getLogger('pyomo.core')


@TransformationFactory.register(
    'mpec.simple_disjunction',
    doc="Disjunctive transformations of complementarity conditions when all variables are non-negative",
)
class MPEC2_Transformation(Transformation):
    def __init__(self):
        super(MPEC2_Transformation, self).__init__()

    def _apply_to(self, instance, **kwds):
        #
        # Setup transformation data
        #
        tdata = instance._transformation_data['mpec.simple_disjunction']
        tdata.compl_cuids = []
        #
        # Iterate over the model finding Complementarity components
        #
        for complementarity in instance.component_objects(
            Complementarity,
            active=True,
            descend_into=(Block, Disjunct),
            sort=SortComponents.deterministic,
        ):
            block = complementarity.parent_block()

            for index in sorted(complementarity.keys()):
                _data = complementarity[index]
                if not _data.active:
                    continue
                #
                _e1 = _data._canonical_expression(_data._args[0])
                _e2 = _data._canonical_expression(_data._args[1])
                if (
                    len(_e1) == 3
                    and len(_e2) == 3
                    and (_e1[0] is None)
                    + (_e1[2] is None)
                    + (_e2[0] is None)
                    + (_e2[2] is None)
                    != 2
                ):
                    raise RuntimeError(
                        "Complementarity condition %s must have exactly two finite bounds"
                        % _data.name
                    )
                if len(_e1) == 3 and _e1[0] is None and _e1[2] is None:
                    #
                    # Swap _e1 and _e2.  The ensures that
                    # only e2 will be an unconstrained expression
                    #
                    _e1, _e2 = _e2, _e1
                if _e2[0] is None and _e2[2] is None:
                    if len(_e1) == 2:
                        _data.c = Constraint(expr=_e1)
                    else:
                        _data.expr1 = Disjunct()
                        _data.expr1.c0 = Constraint(expr=_e1[0] == _e1[1])
                        _data.expr1.c1 = Constraint(expr=_e2[1] >= 0)
                        #
                        _data.expr2 = Disjunct()
                        _data.expr2.c0 = Constraint(expr=_e1[1] == _e1[2])
                        _data.expr2.c1 = Constraint(expr=_e2[1] <= 0)
                        #
                        _data.expr3 = Disjunct()
                        _data.expr3.c0 = Constraint(
                            expr=inequality(_e1[0], _e1[1], _e1[2])
                        )
                        _data.expr3.c1 = Constraint(expr=_e2[1] == 0)
                        _data.complements = Disjunction(
                            expr=(_data.expr1, _data.expr2, _data.expr3)
                        )
                else:
                    if _e1[0] is None:
                        tmp1 = _e1[2] - _e1[1]
                    else:
                        tmp1 = _e1[1] - _e1[0]
                    if _e2[0] is None:
                        tmp2 = _e2[2] - _e2[1]
                    else:
                        tmp2 = _e2[1] - _e2[0]
                    _data.expr1 = Disjunct()
                    _data.expr1.c0 = Constraint(expr=tmp1 >= 0)
                    _data.expr1.c1 = Constraint(expr=tmp2 == 0)
                    #
                    _data.expr2 = Disjunct()
                    _data.expr2.c0 = Constraint(expr=tmp1 == 0)
                    _data.expr2.c1 = Constraint(expr=tmp2 >= 0)
                    #
                    _data.complements = Disjunction(expr=(_data.expr1, _data.expr2))
            tdata.compl_cuids.append(ComponentUID(complementarity))
            block.reclassify_component_type(complementarity, Block)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.errors import PyomoException
from pyomo.core.expr.visitor import ExpressionValueVisitor, nonpyomo_leaf_types
import pyomo.core.expr as EXPR
from pyomo.core.base.objective import Objective


class SolverError(PyomoException):
    """General error raised by Pyomo solver interfaces when processing
    Solver results."""


class NoFeasibleSolutionError(PyomoException):
    default_message = (
        'A feasible solution was not found, so no solution can be loaded. '
        'Please set opt.config.load_solutions=False and check '
        'results.solution_status and '
        'results.incumbent_objective before loading a solution.'
    )


class NoOptimalSolutionError(PyomoException):
    default_message = (
        'Solver did not find the optimal solution. Set '
        'opt.config.raise_exception_on_nonoptimal_result=False to bypass this error.'
    )


class NoSolutionError(PyomoException):
    default_message = (
        'Solution loader does not currently have a valid solution. Please '
        'check results.termination_condition and/or results.solution_status.'
    )


class NoDualsError(PyomoException):
    default_message = (
        'Solver does not currently have valid duals. Please '
        'check results.termination_condition and/or results.solution_status.'
    )


class NoReducedCostsError(PyomoException):
    default_message = (
        'Solver does not currently have valid reduced costs. Please '
        'check results.termination_condition and/or results.solution_status.'
    )


class IncompatibleModelError(PyomoException):
    default_message = (
        'Model is not compatible with the chosen solver. Please check '
        'the model and solver.'
    )


def get_objective(block):
    """
    Get current active objective on a block. If there is more than one active,
    return an error.
    """
    objective = None
    for obj in block.component_data_objects(
        Objective, descend_into=True, active=True, sort=True
    ):
        if objective is not None:
            raise ValueError('Multiple active objectives found.')
        objective = obj
    return objective


class _VarAndNamedExprCollector(ExpressionValueVisitor):
    def __init__(self):
        self.named_expressions = {}
        self.variables = {}
        self.fixed_vars = {}
        self._external_functions = {}

    def visit(self, node, values):
        pass

    def visiting_potential_leaf(self, node):
        if type(node) in nonpyomo_leaf_types:
            return True, None

        if node.is_variable_type():
            self.variables[id(node)] = node
            if node.is_fixed():
                self.fixed_vars[id(node)] = node
            return True, None

        if node.is_named_expression_type():
            self.named_expressions[id(node)] = node
            return False, None

        if isinstance(node, EXPR.ExternalFunctionExpression):
            self._external_functions[id(node)] = node
            return False, None

        if node.is_expression_type():
            return False, None

        return True, None


_visitor = _VarAndNamedExprCollector()


def collect_vars_and_named_exprs(expr):
    _visitor.__init__()
    _visitor.dfs_postorder_stack(expr)
    return (
        list(_visitor.named_expressions.values()),
        list(_visitor.variables.values()),
        list(_visitor.fixed_vars.values()),
        list(_visitor._external_functions.values()),
    )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
import sys
from copy import deepcopy
from pickle import PickleError
from weakref import ref as weakref_ref

import pyomo.common
from pyomo.common import DeveloperError
from pyomo.common.autoslots import AutoSlots, fast_deepcopy
from pyomo.common.collections import OrderedDict
from pyomo.common.deprecation import (
    RenamedClass,
    deprecated,
    deprecation_warning,
    relocated_module_attribute,
)
from pyomo.common.factory import Factory
from pyomo.common.formatting import tabular_writer, StreamIndenter
from pyomo.common.modeling import NOTSET
from pyomo.common.sorting import sorted_robust
from pyomo.core.pyomoobject import PyomoObject
from pyomo.core.base.component_namer import name_repr, index_repr
from pyomo.core.base.enums import SortComponents
from pyomo.core.base.global_set import UnindexedComponent_index
from pyomo.core.base.initializer import PartialInitializer

logger = logging.getLogger('pyomo.core')

relocated_module_attribute(
    'ComponentUID', 'pyomo.core.base.componentuid.ComponentUID', version='5.7.2'
)

_ref_types = {type(None), weakref_ref}
DEFAULT_PPRINT_SORT = SortComponents.ALPHABETICAL | SortComponents.SORTED_INDICES


class ModelComponentFactoryClass(Factory):
    def register(self, doc=None):
        def fn(cls):
            return super(ModelComponentFactoryClass, self).register(cls.__name__, doc)(
                cls
            )

        return fn


ModelComponentFactory = ModelComponentFactoryClass('model component')


def name(component, index=NOTSET, fully_qualified=False, relative_to=None):
    """
    Return a string representation of component for a specific
    index value.
    """
    base = component.getname(fully_qualified=fully_qualified, relative_to=relative_to)
    if index is NOTSET:
        return base
    else:
        if index not in component.index_set():
            raise KeyError(
                "Index %s is not valid for component %s" % (index, component.name)
            )
        return base + index_repr(index)


@deprecated(msg="The cname() function has been renamed to name()", version='5.6.9')
def cname(*args, **kwds):
    return name(*args, **kwds)


class CloneError(pyomo.common.errors.PyomoException):
    pass


class ComponentBase(PyomoObject):
    """A base class for :py:class:`Component` and :py:class:`ComponentData`

    This class defines some fundamental methods and properties that are
    expected for all Component-like objects.  They are centralized here
    to avoid repeated code in the :py:class:`Component` and
    :py:class:`ComponentData` classes.

    """

    __slots__ = ()

    _PPRINT_INDENT = "    "

    def is_component_type(self):
        """Return True if this class is a Pyomo component"""
        return True

    def __deepcopy__(self, memo):
        # The problem we are addressing is when we want to clone a
        # sub-block in a model.  In that case, the block can have
        # references to both child components and to external
        # ComponentData (mostly through expressions pointing to Vars
        # and Params outside this block).  For everything stored beneath
        # this block, we want to clone the Component (and all
        # corresponding ComponentData objects).  But for everything
        # stored outside this Block, we want to do a simple shallow
        # copy.
        #
        # Nominally, expressions only point to ComponentData
        # derivatives.  However, with the development of Expression
        # Templates (and the corresponding _GetItemExpression object),
        # expressions can refer to container (non-Simple) components, so
        # we need to override __deepcopy__ for both Component and
        # ComponentData (so we put it here on ComponentBase).
        #
        if '__block_scope__' in memo:
            _scope = memo['__block_scope__']
            _new = None
            tmp = self.parent_block()
            # "Floating" components should be in scope by default (we
            # will handle 'global' components like GlobalSets in the
            # components).  This ensures that things like set operators
            # on Abstract set objects are correctly cloned.  For
            # example, consider an abstract indexed model component
            # whose domain is specified by a Set expression:
            #
            #   def x_init(m,i):
            #       if i == 2:
            #           return Set.Skip
            #       else:
            #           return []
            #   m.x = Set( [1,2],
            #              domain={1: m.A*m.B, 2: m.A*m.A},
            #              initialize=x_init )
            #
            # We do not want to automatically add all the Set operators
            # to the model at declaration time, as m.x[2] is never
            # actually created.  Plus, doing so would require complex
            # parsing of the initializers.  BUT, we need to ensure that
            # the operators are deepcopied, otherwise when the model is
            # cloned before construction the operators will still refer
            # to the sets on the original abstract model (in particular,
            # the Set x will have an unknown dimen).
            #
            # The solution is to automatically clone all floating
            # components, except for Models (i.e., top-level BlockData
            # have no parent and technically "float")
            _in_scope = tmp is None and self is not self.model()
            #
            # Note: normally we would need to check that tmp does not
            # end up being None.  However, since clone() inserts
            # id(None) into the __block_scope__ dictionary, we are safe
            while id(tmp) not in _scope:
                _new = (_new, id(tmp))
                tmp = tmp.parent_block()
            _in_scope |= _scope[id(tmp)]

            # Remember whether all newly-encountered blocks are in or
            # out of scope (prevent duplicate work)
            while _new is not None:
                _new, _id = _new
                _scope[_id] = _in_scope

            # Note: there is an edge case when cloning a block: the
            # initial call to deepcopy (on the target block) has
            # __block_scope__ defined, however, the parent block of self
            # is either None, or is (by definition) out of scope.  So we
            # will check that id(self) is not in __block_scope__: if it
            # is, then this is the top-level block and we need to do the
            # normal deepcopy.  We defer this check until now for
            # efficiency reasons because we expect that (for sane models)
            # the bulk of the components we will encounter will be *in*
            # scope.
            if not _in_scope and id(self) not in _scope:
                # component is out-of-scope.  shallow copy only
                memo[id(self)] = self
                return self
        #
        # deepcopy() is an inherently recursive operation.  This can
        # cause problems for highly interconnected Pyomo models (for
        # example, a time linked model where each time block has a
        # linking constraint [in the time block] to the next / previous
        # block).  This would effectively put the entire time horizon on
        # the stack.  To avoid this, we will leverage the useful
        # knowledge that all component references point to other
        # components / component datas, and NOT to attributes on the
        # components/datas.  So, if we can first go through and stub in
        # all the objects that we will need to populate, and then go
        # through and deepcopy them, we can unroll the vast
        # majority of the recursion.
        #
        component_list = []
        self._create_objects_for_deepcopy(memo, component_list)
        #
        # Note that self is now the first element of component_list
        #
        # Now that we have created (but not populated) all the
        # components that we expect to need in the memo, we can go
        # through and populate all the components.
        #
        # The component_list is roughly in declaration order.  This
        # means that it should be relatively safe to clone the contents
        # in the same order.
        #
        for comp, new in component_list:
            comp.__deepcopy_state__(memo, new)
        return memo[id(self)]

    def _create_objects_for_deepcopy(self, memo, component_list):
        _new = self.__class__.__new__(self.__class__)
        _ans = memo.setdefault(id(self), _new)
        if _ans is _new:
            component_list.append((self, _new))
        return _ans

    def __deepcopy_field__(self, value, memo, slot_name):
        saved_memo = len(memo)
        try:
            return fast_deepcopy(value, memo)
        except CloneError:
            raise
        except:
            # remove entries added to the memo
            for _ in range(len(memo) - saved_memo):
                memo.popitem()
            # warn the user
            if '__block_scope__' not in memo:
                logger.warning(
                    "Uncopyable field encountered when deep "
                    "copying Pyomo components outside the scope of "
                    "Block.clone().  There is a distinct possibility "
                    "that the new copy is not complete.  To avoid "
                    "this situation, please use Block.clone()"
                )
            if self.model() is self:
                what = 'Model'
            else:
                what = 'Component'
            logger.error(
                "Unable to clone Pyomo component attribute.\n"
                "%s '%s' contains an uncopyable field '%s' (%s).  "
                "Setting field to `None` on new object"
                % (what, self.name, slot_name, type(value))
            )
            # If this is an abstract model, then we are probably
            # in the middle of create_instance, and the model
            # that will eventually become the concrete model is
            # missing initialization data.  This is an
            # exceptional event worthy of a stronger (and more
            # informative) error.
            if not self.parent_component()._constructed:
                raise CloneError(
                    "Uncopyable attribute (%s) encountered when "
                    "cloning component %s on an abstract block.  "
                    "The resulting instance is therefore "
                    "missing data from the original abstract model "
                    "and likely will not construct correctly.  "
                    "Consider changing how you initialize this "
                    "component or using a ConcreteModel." % (slot_name, self.name)
                )
        # Drop the offending field value.  The user has been warned.
        return None

    @deprecated(
        """The cname() method has been renamed to getname().
    The preferred method of obtaining a component name is to use the
    .name property, which returns the fully qualified component name.
    The .local_name property will return the component name only within
    the context of the immediate parent container.""",
        version='5.0',
    )
    def cname(self, *args, **kwds):
        return self.getname(*args, **kwds)

    def pprint(self, ostream=None, verbose=False, prefix="", sort=NOTSET):
        """Print component information

        Note that this method is generally only reachable through
        ComponentData objects in an IndexedComponent container.
        Components, including unindexed Component derivatives and both
        scalar and indexed IndexedComponent derivatives will see
        :py:meth:`Component.pprint()`
        """
        comp = self.parent_component()
        _attr, _data, _header, _fcn = comp._pprint()
        if isinstance(type(_data), str):
            # If the component _pprint only returned a pre-formatted
            # result, then we have no way to only emit the information
            # for this _data object.
            _name = comp.local_name
        else:
            # restrict output to only this data object
            _data = lambda _sort: iter(((self.index(), self),))
            _name = "{Member of %s}" % (comp.local_name,)
        self._pprint_base_impl(
            ostream,
            verbose,
            prefix,
            sort,
            _name,
            comp.doc,
            comp.is_constructed(),
            _attr,
            _data,
            _header,
            _fcn,
        )

    @property
    def name(self):
        """Get the fully qualified component name."""
        return self.getname(fully_qualified=True)

    # Adding a setter here to help users adapt to the new
    # setting. The .name attribute is now ._name. It should
    # never be assigned to by user code.
    @name.setter
    def name(self, val):
        raise ValueError(
            "The .name attribute is now a property method "
            "that returns the fully qualified component name. "
            "Assignment is not allowed."
        )

    @property
    def local_name(self):
        """Get the component name only within the context of
        the immediate parent container."""
        return self.getname(fully_qualified=False)

    @property
    def active(self):
        """Return the active attribute"""
        # Normal components cannot be deactivated
        return True

    @active.setter
    def active(self, value):
        """Set the active attribute to the given value"""
        raise AttributeError(
            "Setting the 'active' flag on a component that does not "
            "support deactivation is not allowed."
        )

    def _pprint_base_impl(
        self,
        ostream,
        verbose,
        prefix,
        sort,
        _name,
        _doc,
        _constructed,
        _attr,
        _data,
        _header,
        _fcn,
    ):
        if ostream is None:
            ostream = sys.stdout
        if prefix:
            ostream = StreamIndenter(ostream, prefix)

        if sort is NOTSET:
            sort = DEFAULT_PPRINT_SORT
        sort = SortComponents(sort)

        # FIXME: HACK for backwards compatibility with suppressing the
        # header for the top block
        if not _attr and self.parent_block() is None:
            _name = ''

        if self.is_reference():
            _attr = list(_attr) if _attr else []
            _attr.append(('ReferenceTo', self.referent))

        if _name:
            ostream.write(_name + " : ")
        if _doc:
            ostream.write(_doc + '\n' + self._PPRINT_INDENT)
        if _attr:
            ostream.write(", ".join("%s=%s" % (k, v) for k, v in _attr))
        if _attr or _name or _doc:
            ostream.write("\n")
            # We only indent everything if we printed the header
            ostream = StreamIndenter(ostream, self._PPRINT_INDENT)

        if not _constructed:
            # HACK: for backwards compatibility, Abstract blocks will
            # still print their assigned components.  Should we instead
            # always pprint unconstructed components (possibly
            # suppressing the table header if the table is empty)?
            if self.parent_block() is not None:
                ostream.write("Not constructed\n")
                return

        if type(_fcn) is tuple:
            # Exception to the standard formatter case: with two
            # callbacks, we will use the first to generate the normal
            # table, then call the second callback for each data.
            # Currently only used by Complimentarity (which should be
            # refactored to remove the need for this edge case)
            _fcn, _fcn2 = _fcn
        else:
            _fcn2 = None

        if hasattr(_data, '__call__'):
            _data = _data(sort)

        if _header is not None:
            # This is a standard component, where all the component
            # information is printed in a single table
            if _fcn2 is not None:
                _data = list(_data)
            tabular_writer(ostream, '', _data, _header, _fcn)
            if _fcn2 is not None:
                for _key, _val in _data:
                    _fcn2(ostream, sort, _key, _val)
        elif _fcn is not None:
            # This is a non-standard component where we will not
            # generate a table at all, and instead defer all formatting
            # / printing to the callback.  This is primarily used by
            # Blocks (and block-like things)
            for _key, _val in _data:
                _fcn(ostream, sort, _key, _val)
        elif _data is not None:
            # Catch all for everything else: assume that _pprint()
            # returned a formatted string.
            ostream.write(_data)


class _ComponentBase(metaclass=RenamedClass):
    __renamed__new_class__ = ComponentBase
    __renamed__version__ = '6.7.2'


class Component(ComponentBase):
    """
    This is the base class for all Pyomo modeling components.

    Parameters
    ----------
    ctype : type
        The class type for the derived subclass

    doc : str
        A text string describing this component

    name : str
        A name for this component

    Attributes
    ----------
    doc : str
        A text string describing this component

    """

    __autoslot_mappers__ = {'_parent': AutoSlots.weakref_mapper}

    def __init__(self, **kwds):
        #
        # Get arguments
        #
        self._ctype = kwds.pop('ctype', None)
        self.doc = kwds.pop('doc', None)
        self._name = kwds.pop('name', None)
        if kwds:
            # If there are leftover keywords, and the component has a
            # rule, pass those keywords on to the rule
            if getattr(self, '_rule', None) is not None:
                self._rule = PartialInitializer(self._rule, **kwds)
            else:
                raise ValueError(
                    "Unexpected keyword options found while constructing '%s':\n\t%s"
                    % (type(self).__name__, ','.join(sorted(kwds.keys())))
                )
        #
        # Verify that ctype has been specified.
        #
        if self._ctype is None:
            raise DeveloperError(
                "Must specify a component type for class %s!" % (type(self).__name__,)
            )
        #
        self._constructed = False
        self._parent = None  # Must be a weakref

    @property
    def ctype(self):
        """Return the class type for this component"""
        return self._ctype

    @deprecated(
        "Component.type() method has been replaced by the .ctype property.",
        version='5.7',
    )
    def type(self):
        """Return the class type for this component"""
        return self.ctype

    def construct(self, data=None):  # pragma:nocover
        """API definition for constructing components"""
        pass

    def is_constructed(self):  # pragma:nocover
        """Return True if this class has been constructed"""
        return self._constructed

    def reconstruct(self, data=None):
        """REMOVED: reconstruct() was removed in Pyomo 6.0.

        Re-constructing model components was fragile and did not
        correctly update instances of the component used in other
        components or contexts (this was particularly problemmatic for
        Var, Param, and Set).  Users who wish to reproduce the old
        behavior of reconstruct(), are comfortable manipulating
        non-public interfaces, and who take the time to verify that the
        correct thing happens to their model can approximate the old
        behavior of reconstruct with:

            component.clear()
            component._constructed = False
            component.construct()

        """
        raise AttributeError(self.reconstruct.__doc__)

    def valid_model_component(self):
        """Return True if this can be used as a model component."""
        return True

    def pprint(self, ostream=None, verbose=False, prefix="", sort=NOTSET):
        """Print component information"""
        self._pprint_base_impl(
            ostream,
            verbose,
            prefix,
            sort,
            self.local_name,
            self.doc,
            self.is_constructed(),
            *self._pprint(),
        )

    def display(self, ostream=None, verbose=False, prefix=""):
        self.pprint(ostream=ostream, prefix=prefix)

    def parent_component(self):
        """Returns the component associated with this object."""
        return self

    def parent_block(self):
        """Returns the parent of this object."""
        if self._parent is None:
            return None
        else:
            return self._parent()

    def model(self):
        """Returns the model associated with this object."""
        # This is a re-implementation of Component.parent_block(),
        # duplicated for effficiency to avoid the method call
        if self._parent is None:
            return None
        ans = self._parent()

        if ans is None:
            return None
        # Defer to the (simple) block's model() method to walk up the
        # hierarchy. This is because the top-level block can be a model,
        # but nothing else (e.g., calling model() on a Var not attached
        # to a model should return None, but calling model() on a Block
        # not attached to anything else should return the Block)
        return ans.model()

    def root_block(self):
        """Return self.model()"""
        return self.model()

    def __str__(self):
        """Return the component name"""
        return self.name

    def getname(self, fully_qualified=False, name_buffer=None, relative_to=None):
        """Returns the component name associated with this object.

        Parameters
        ----------
        fully_qualified: bool
            Generate full name from nested block names

        relative_to: Block
            Generate fully_qualified names relative to the specified block.
        """
        local_name = self._name
        if local_name is None:
            local_name = type(self).__name__
        if fully_qualified:
            pb = self.parent_block()
            if relative_to is None:
                relative_to = self.model()
            if pb is not None and pb is not relative_to:
                ans = (
                    pb.getname(fully_qualified, name_buffer, relative_to)
                    + "."
                    + name_repr(local_name)
                )
            elif pb is None and relative_to != self.model():
                raise RuntimeError(
                    "The relative_to argument was specified but not found "
                    "in the block hierarchy: %s" % str(relative_to)
                )
            else:
                ans = name_repr(local_name)
        else:
            # Note: we want "getattr(x.parent_block(), x.local_name) == x"
            # so we do not want to call _safe_name_str, as that could
            # add quotes or otherwise escape the string.
            ans = local_name
        if name_buffer is not None:
            deprecation_warning(
                "The 'name_buffer' argument to getname is deprecated. "
                "The functionality is no longer necessary since getting names "
                "is no longer a quadratic operation. Additionally, note that "
                "use of this argument poses risks if the buffer contains "
                "names relative to different Blocks in the model hierarchy or "
                "a mixture of local and fully_qualified names.",
                version='6.4.1',
            )
            name_buffer[id(self)] = ans
        return ans

    @property
    def name(self):
        """Get the fully qualified component name."""
        return self.getname(fully_qualified=True)

    # Allow setting a component's name if it is not owned by a parent
    # block (this supports, e.g., naming a model)
    @name.setter
    def name(self, val):
        if self.parent_block() is None:
            self._name = val
        else:
            raise ValueError(
                "The .name attribute is not settable when the component "
                "is assigned to a Block.\nTriggered by attempting to set "
                "component '%s' to name '%s'" % (self.name, val)
            )

    def is_indexed(self):
        """Return true if this component is indexed"""
        return False

    def clear_suffix_value(self, suffix_or_name, expand=True):
        """Clear the suffix value for this component data"""
        if isinstance(suffix_or_name, str):
            import pyomo.core.base.suffix

            for name_, suffix_ in pyomo.core.base.suffix.active_suffix_generator(
                self.model()
            ):
                if suffix_or_name == name_:
                    suffix_.clear_value(self, expand=expand)
                    break
        else:
            suffix_or_name.clear_value(self, expand=expand)

    def set_suffix_value(self, suffix_or_name, value, expand=True):
        """Set the suffix value for this component data"""
        if isinstance(suffix_or_name, str):
            import pyomo.core.base.suffix

            for name_, suffix_ in pyomo.core.base.suffix.active_suffix_generator(
                self.model()
            ):
                if suffix_or_name == name_:
                    suffix_.set_value(self, value, expand=expand)
                    break
        else:
            suffix_or_name.set_value(self, value, expand=expand)

    def get_suffix_value(self, suffix_or_name, default=None):
        """Get the suffix value for this component data"""
        if isinstance(suffix_or_name, str):
            import pyomo.core.base.suffix

            for name_, suffix_ in pyomo.core.base.suffix.active_suffix_generator(
                self.model()
            ):
                if suffix_or_name == name_:
                    return suffix_.get(self, default)
        else:
            return suffix_or_name.get(self, default)

    def _pop_from_kwargs(self, name, kwargs, namelist, notset=None):
        args = [
            arg
            for arg in (kwargs.pop(name, notset) for name in namelist)
            if arg is not notset
        ]
        if len(args) == 1:
            return args[0]
        elif not args:
            return notset
        else:
            argnames = "%s%s '%s='" % (
                ', '.join("'%s='" % _ for _ in namelist[:-1]),
                ',' if len(namelist) > 2 else '',
                namelist[-1],
            )
            raise ValueError(
                "Duplicate initialization: %s() only accepts one of %s"
                % (name, argnames)
            )


class ActiveComponent(Component):
    """A Component that makes semantic sense to activate or deactivate
    in a model.

    Private class attributes:
        _active         A boolean that is true if this component will be
                            used in model operations
    """

    def __init__(self, **kwds):
        self._active = True
        super(ActiveComponent, self).__init__(**kwds)

    @property
    def active(self):
        """Return the active attribute"""
        return self._active

    @active.setter
    def active(self, value):
        """Set the active attribute to the given value"""
        raise AttributeError("Assignment not allowed. Use the (de)activate methods.")

    def activate(self):
        """Set the active attribute to True"""
        self._active = True

    def deactivate(self):
        """Set the active attribute to False"""
        self._active = False


class ComponentData(ComponentBase):
    """
    This is the base class for the component data used
    in Pyomo modeling components.  Subclasses of ComponentData are
    used in indexed components, and this class assumes that indexed
    components are subclasses of IndexedComponent.

    Constructor arguments:
        owner           The component that owns this data object

    Private class attributes:
        _component      A weakref to the component that owns this data object
        _index          The index of this data object
    """

    __slots__ = ('_component', '_index', '__weakref__')
    __autoslot_mappers__ = {'_component': AutoSlots.weakref_mapper}

    # NOTE: This constructor is in-lined in the constructors for the following
    # classes: BooleanVarData, ConnectorData, ConstraintData,
    # ExpressionData, LogicalConstraintData,
    # LogicalConstraintData, ObjectiveData,
    # ParamData,VarData, BooleanVarData, DisjunctionData,
    # ArcData, PortData, _LinearConstraintData, and
    # _LinearMatrixConstraintData. Changes made here need to be made in those
    # constructors as well!
    def __init__(self, component):
        #
        # ComponentData objects are typically *private* objects for
        # indexed / sparse indexed components.  As such, the (derived)
        # class needs to make sure that the owning component is *always*
        # passed as the owner (and that owner is never None).  Not validating
        # this assumption is significantly faster.
        #
        self._component = weakref_ref(component)
        self._index = NOTSET

    @property
    def ctype(self):
        """Return the class type for this component"""
        _parent = self.parent_component()
        if _parent is None:
            return None
        return _parent._ctype

    @deprecated(
        "Component.type() method has been replaced by the .ctype property.",
        version='5.7',
    )
    def type(self):
        """Return the class type for this component"""
        return self.ctype

    def parent_component(self):
        """Returns the component associated with this object."""
        if self._component is None:
            return None
        return self._component()

    def parent_block(self):
        """Return the parent of the component that owns this data."""
        # This is a re-implementation of parent_component(), duplicated
        # for effficiency to avoid the method call
        if self._component is None:
            return None
        comp = self._component()

        # This is a re-implementation of Component.parent_block(),
        # duplicated for effficiency to avoid the method call
        if comp._parent is None:
            return None
        return comp._parent()

    def model(self):
        """Return the model of the component that owns this data."""
        ans = self.parent_block()
        if ans is None:
            return None
        # Defer to the (simple) block's model() method to walk up the
        # hierarchy. This is because the top-level block can be a model,
        # but nothing else (e.g., calling model() on a Var not attached
        # to a model should return None, but calling model() on a Block
        # not attached to anything else should return the Block)
        return ans.model()

    def index(self):
        """
        Returns the index of this ComponentData instance relative
        to the parent component index set. None is returned if
        this instance does not have a parent component, or if
        - for some unknown reason - this instance does not belong
        to the parent component's index set.
        """
        try:
            if self._component()[self._index] is self:
                return self._index
        except:
            pass
        if self._index is NOTSET:
            return self._index
        parent = self.parent_component()
        if parent is None:
            return self._index
        # This error message is a bit goofy, but we can't call self.name
        # here--it's an infinite loop!
        raise DeveloperError(
            "The '_data' dictionary and '_index' attribute are out of "
            "sync for indexed %s '%s': The %s entry in the '_data' "
            "dictionary does not map back to this component data object."
            % (parent.ctype.__name__, parent.name, self._index)
        )

    def __str__(self):
        """Return a string with the component name and index"""
        return self.name

    def getname(self, fully_qualified=False, name_buffer=None, relative_to=None):
        """Return a string with the component name and index"""
        # NOTE: There are bugs with name buffers if a user always gives the same
        # dictionary but switches from fully-qualified to local names or changes
        # the component the name is relative to. We will simply deprecate the
        # buffer in a future PR, but for now we will leave this method so it
        # behaves the same when a buffer is given, and, in the absence of a
        # buffer it will construct the name using the index (woohoo!)

        #
        # Using the buffer, which is a dictionary:  id -> string
        #
        if name_buffer is not None:
            deprecation_warning(
                "The 'name_buffer' argument to getname is deprecated. "
                "The functionality is no longer necessary since getting names "
                "is no longer a quadratic operation. Additionally, note that "
                "use of this argument poses risks if the buffer contains "
                "names relative to different Blocks in the model hierarchy or "
                "a mixture of local and fully_qualified names.",
                version='6.4.1',
            )
            if id(self) in name_buffer:
                # Return the name if it is in the buffer
                return name_buffer[id(self)]

        c = self.parent_component()
        if c is self:
            #
            # This is a scalar component, so call the
            # Component.getname() method
            #
            return super(ComponentData, self).getname(
                fully_qualified, name_buffer, relative_to
            )
        elif c is not None:
            #
            # Get the name of the parent component
            #
            base = c.getname(fully_qualified, name_buffer, relative_to)
        else:
            #
            # Defensive: this is a ComponentData without a valid
            # parent_component.  As this usually occurs when handling
            # exceptions during model construction, we need to ensure
            # that this method doesn't itself raise another exception.
            #
            return '[Unattached %s]' % (type(self).__name__,)

        if name_buffer is not None:
            # Iterate through the dictionary and generate all names in
            # the buffer
            for idx, obj in c.items():
                name_buffer[id(obj)] = base + index_repr(idx)
            if id(self) in name_buffer:
                # Return the name if it is in the buffer
                return name_buffer[id(self)]
        else:
            #
            # No buffer, we can do what we are going to do all the time after we
            # deprecate the buffer.
            #
            return base + index_repr(self.index())
        #
        raise RuntimeError(
            "Fatal error: cannot find the component data in "
            "the owning component's _data dictionary."
        )

    def is_indexed(self):
        """Return true if this component is indexed"""
        return False

    def clear_suffix_value(self, suffix_or_name, expand=True):
        """Set the suffix value for this component data"""
        if isinstance(suffix_or_name, str):
            import pyomo.core.base.suffix

            for name_, suffix_ in pyomo.core.base.suffix.active_suffix_generator(
                self.model()
            ):
                if suffix_or_name == name_:
                    suffix_.clear_value(self, expand=expand)
                    break
        else:
            suffix_or_name.clear_value(self, expand=expand)

    def set_suffix_value(self, suffix_or_name, value, expand=True):
        """Set the suffix value for this component data"""
        if isinstance(suffix_or_name, str):
            import pyomo.core.base.suffix

            for name_, suffix_ in pyomo.core.base.suffix.active_suffix_generator(
                self.model()
            ):
                if suffix_or_name == name_:
                    suffix_.set_value(self, value, expand=expand)
                    break
        else:
            suffix_or_name.set_value(self, value, expand=expand)

    def get_suffix_value(self, suffix_or_name, default=None):
        """Get the suffix value for this component data"""
        if isinstance(suffix_or_name, str):
            import pyomo.core.base.suffix

            for name_, suffix_ in pyomo.core.base.suffix.active_suffix_generator(
                self.model()
            ):
                if suffix_or_name == name_:
                    return suffix_.get(self, default)
        else:
            return suffix_or_name.get(self, default)


class ActiveComponentData(ComponentData):
    """
    This is the base class for the component data used
    in Pyomo modeling components that can be activated and
    deactivated.

    It's possible to end up in a state where the parent Component
    has _active=True but all ComponentData have _active=False. This
    seems like a reasonable state, though we cannot easily detect
    this situation.  The important thing to avoid is the situation
    where one or more ComponentData are active, but the parent
    Component claims active=False. This class structure is designed
    to prevent this situation.

    Constructor arguments:
        owner           The component that owns this data object

    Private class attributes:
        _component      A weakref to the component that owns this data object
        _index          The index of this data object
        _active         A boolean that indicates whether this data is active
    """

    __slots__ = ('_active',)

    def __init__(self, component):
        super(ActiveComponentData, self).__init__(component)
        self._active = True

    @property
    def active(self):
        """Return the active attribute"""
        return self._active

    @active.setter
    def active(self, value):
        """Set the active attribute to a specified value."""
        raise AttributeError("Assignment not allowed. Use the (de)activate method")

    def activate(self):
        """Set the active attribute to True"""
        self._active = self.parent_component()._active = True

    def deactivate(self):
        """Set the active attribute to False"""
        self._active = False



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
from pyomo.common.dependencies import attempt_import

np, np_available = attempt_import('numpy', minimum_version='1.13.0')
scipy, scipy_available = attempt_import('scipy.sparse')
mumps, mumps_available = attempt_import('mumps')
if not np_available or not scipy_available:
    raise unittest.SkipTest('numpy and scipy are needed for interior point tests')
import numpy as np
from scipy.sparse import coo_matrix, tril
from pyomo.contrib import interior_point as ip
from pyomo.contrib.pynumero.linalg.base import LinearSolverStatus

if scipy_available:
    from pyomo.contrib.interior_point.linalg.scipy_interface import ScipyInterface
if mumps_available:
    from pyomo.contrib.interior_point.linalg.mumps_interface import MumpsInterface
from pyomo.contrib.pynumero.linalg.ma27 import MA27Interface

ma27_available = MA27Interface.available()
if ma27_available:
    from pyomo.contrib.interior_point.linalg.ma27_interface import (
        InteriorPointMA27Interface,
    )


def get_base_matrix(use_tril):
    if use_tril:
        row = [0, 1, 1, 2, 2]
        col = [0, 0, 1, 0, 2]
        data = [1, 7, 4, 3, 6]
    else:
        row = [0, 0, 0, 1, 1, 2, 2]
        col = [0, 1, 2, 0, 1, 0, 2]
        data = [1, 7, 3, 7, 4, 3, 6]
    mat = coo_matrix((data, (row, col)), shape=(3, 3), dtype=np.double)
    return mat


def get_base_matrix_wrong_order(use_tril):
    if use_tril:
        row = [1, 0, 1, 2, 2]
        col = [0, 0, 1, 0, 2]
        data = [7, 1, 4, 3, 6]
    else:
        row = [1, 0, 0, 0, 1, 2, 2]
        col = [0, 1, 2, 0, 1, 0, 2]
        data = [7, 7, 3, 1, 4, 3, 6]
    mat = coo_matrix((data, (row, col)), shape=(3, 3), dtype=np.double)
    return mat


class TestTrilBehavior(unittest.TestCase):
    """
    Some of the other tests in this file depend on
    the behavior of tril that is tested in this
    test, namely the tests in TestWrongNonzeroOrdering.
    """

    def test_tril_behavior(self):
        mat = get_base_matrix(use_tril=True)
        mat2 = tril(mat)
        self.assertTrue(np.all(mat.row == mat2.row))
        self.assertTrue(np.all(mat.col == mat2.col))
        self.assertTrue(np.allclose(mat.data, mat2.data))

        mat = get_base_matrix_wrong_order(use_tril=True)
        self.assertFalse(np.all(mat.row == mat2.row))
        self.assertFalse(np.allclose(mat.data, mat2.data))
        mat2 = tril(mat)
        self.assertTrue(np.all(mat.row == mat2.row))
        self.assertTrue(np.all(mat.col == mat2.col))
        self.assertTrue(np.allclose(mat.data, mat2.data))


class TestLinearSolvers(unittest.TestCase):
    def _test_linear_solvers(self, solver):
        mat = get_base_matrix(use_tril=False)
        zero_mat = mat.copy()
        zero_mat.data.fill(0)
        stat = solver.do_symbolic_factorization(zero_mat)
        self.assertEqual(stat.status, LinearSolverStatus.successful)
        stat = solver.do_numeric_factorization(mat)
        self.assertEqual(stat.status, LinearSolverStatus.successful)
        x_true = np.array([1, 2, 3], dtype=np.double)
        rhs = mat * x_true
        x, res = solver.do_back_solve(rhs)
        self.assertTrue(np.allclose(x, x_true))
        x_true = np.array([4, 2, 3], dtype=np.double)
        rhs = mat * x_true
        x, res = solver.do_back_solve(rhs)
        self.assertTrue(np.allclose(x, x_true))

    @unittest.skipIf(
        not scipy_available, 'scipy is needed for interior point scipy tests'
    )
    def test_scipy(self):
        solver = ScipyInterface()
        self._test_linear_solvers(solver)

    @unittest.skipIf(
        not mumps_available, 'mumps is needed for interior point mumps tests'
    )
    def test_mumps(self):
        solver = MumpsInterface()
        self._test_linear_solvers(solver)

    @unittest.skipIf(not ma27_available, 'MA27 is needed for interior point MA27 tests')
    def test_ma27(self):
        solver = InteriorPointMA27Interface()
        self._test_linear_solvers(solver)


@unittest.skip('This does not work yet')
class TestWrongNonzeroOrdering(unittest.TestCase):
    def _test_solvers(self, solver, use_tril):
        mat = get_base_matrix(use_tril=use_tril)
        wrong_order_mat = get_base_matrix_wrong_order(use_tril=use_tril)
        stat = solver.do_symbolic_factorization(mat)
        stat = solver.do_numeric_factorization(wrong_order_mat)
        x_true = np.array([1, 2, 3], dtype=np.double)
        rhs = mat * x_true
        x, res = solver.do_back_solve(rhs)
        self.assertTrue(np.allclose(x, x_true))

    @unittest.skipIf(
        not scipy_available, 'scipy is needed for interior point scipy tests'
    )
    def test_scipy(self):
        solver = ScipyInterface()
        self._test_solvers(solver, use_tril=False)

    @unittest.skipIf(
        not mumps_available, 'mumps is needed for interior point mumps tests'
    )
    def test_mumps(self):
        solver = MumpsInterface()
        self._test_solvers(solver, use_tril=True)

    @unittest.skipIf(not ma27_available, 'MA27 is needed for interior point MA27 tests')
    def test_ma27(self):
        solver = InteriorPointMA27Interface()
        self._test_solvers(solver, use_tril=True)
